from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import dkHS_dxhi00, d2kHS_dxhi00, d3kHS_dxhi00
from .monomer_aux import dkHS_dx_dxhi00, dkHS_dx_dxhi002

from .a1sB_monomer import da1sB_dxhi00_eval, d2a1sB_dxhi00_eval
from .a1sB_monomer import d3a1sB_dxhi00_eval
from .a1sB_monomer import da1sB_dx_dxhi00_dxxhi_eval
from .a1sB_monomer import da1sB_dx_d2xhi00_dxxhi_eval

from .a1_monomer import a1, da1_dxhi00, d2a1_dxhi00,  da1_dx, da1_dxxhi
from .a2_monomer import a2, da2_dxhi00, d2a2_dxhi00,  da2_dx, da2_dxxhi
from .ahs_monomer import ahs, dahs_dxhi00, d2ahs_dxhi00,  dahs_dx, dahs_dxxhi
from .a3_monomer import a3, da3_dxhi00, d2a3_dxhi00,  da3_dx, da3_dxxhi

from .ghs_chain import gdHS, dgdHS_dxhi00, d2gdHS_dxhi00, dgdHS_dx, dgdHS_dxxhi
from .g2mca_chain import g2mca, dg2mca_dxhi00, d2g2mca_dxhi00, dg2mca_dx
from .g2mca_chain import dg2mca_dxxhi
from .gammac_chain import gammac, dgammac_dxhi00, d2gammac_dxhi00, dgammac_dx
from .gammac_chain import dgammac_dxxhi
from .g1sigma_chain import g1sigma, dg1sigma_dxhi00, d2g1sigma_dxhi00
from .g1sigma_chain import dg1sigma_dx, dg1sigma_dxxhi
from .a2new_chain import da2new_dxhi00, d2a2new_dxhi00, d3a2new_dxhi00
from .a2new_chain import da2new_dx_dxhi00, da2new_dxxhi_dxhi00
from .lngmie_chain import lngmie, dlngmie_dxhi00, d2lngmie_dxhi00
from .lngmie_chain import dlngmie_dx, dlngmie_dxxhi

from .association_aux import Iab, dIab_drho, d2Iab_drho
from .association_aux import Xass_solver, dXass_drho, d2Xass_drho, dXass_dx
from .association_aux import dIab_dx, dIab_dxrho, CIJ_matrix

from .polarGV import Apolar, dApolar_drho, d2Apolar_drho
from .polarGV import dApolar_dx, dApolar_dxrho


def xhi_eval(xhi00, xs, xmi, xm, di03):
    xhi = xhi00 * xm * np.matmul(xs, di03)
    dxhi_dxhi00 = np.matmul(xmi, di03)
    dxhi_dxhi00[0] = xm
    return xhi, dxhi_dxhi00


def xhix_eval(xhi00, xs, xm, dij3):
    aux1 = xs * dij3
    aux2 = np.dot(xs, aux1)
    aux3 = aux2.sum()
    dxhix_dxhi00 = xm * aux3
    xhix = xhi00 * dxhix_dxhi00
    return xhix, dxhix_dxhi00


def dxhi_dx_eval(xhi00, xs, xmi, xm, ms, di03):
    xhi = xhi00 * xm * np.matmul(xs, di03)
    dxhi_dxhi00 = np.matmul(xmi, di03)
    dxhi_dxhi00[0] = xm
    dxhi_dx = (xhi00 * di03.T * ms)
    return xhi, dxhi_dxhi00, dxhi_dx


def dxhix_dx_eval(xhi00, xs, dxs_dx, xm, ms, dij3):
    aux1 = xs * dij3
    aux2 = np.dot(xs, aux1)
    aux3 = aux2.sum()
    dxhix_dxhi00 = xm * aux3
    xhix = xhi00 * dxhix_dxhi00
    # suma1 = 2*np.sum(dxs_dx.T@aux1, axis=1)
    suma1 = 2*np.sum(dxs_dx@aux1, axis=1)
    dxhix_dx_dxhi00 = (ms * aux3 + xm * suma1)
    dxhix_dx = xhi00 * dxhix_dx_dxhi00
    return xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00


def ares(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3 = temp_aux[:10]
    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij = temp_aux[10:14]
    beps, beps2, a1vdw_cte, x0i_matrix, tetha = temp_aux[14:19]
    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii = temp_aux[19:25]

    dxhi00_drho = self.dxhi00_drho
    diag_index = self.diag_index

    xmi = x * self.ms
    xm = np.sum(xmi)
    # Eq A8
    xs = xmi / xm

    # Defining xhi0 without x dependence
    xhi00 = dxhi00_drho * rho

    # Eq A7
    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs, xmi, xm, di03)
    # xhi x Eq A13
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs, xm, dij3)
    # xhi m Eq A23
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs, xm, self.sigmaij3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3]])

    khs, dkhs = dkHS_dxhi00(xhix, dxhix_dxhi00)

    da1, da2 = da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij,
                                 J_lambdasij, self.cctesij, a1vdwij,
                                 a1vdw_cteij, dxhix_dxhi00)

    suma1_monomer = self.Cij * np.sum(da1 * x0_a1, axis=1)
    suma2_monomer = self.Cij2 * np.sum(da2 * x0_a2, axis=1)

    # Monomer evaluation
    a1ij = suma1_monomer[0]
    a2ij = suma2_monomer[0]
    aHS = ahs(xhi)
    a1m = a1(xs, a1ij)
    a2m = a2(xs, khs, xhixm, a2ij, epsij, self.f1, self.f2, self.f3)
    a3m = a3(xs, xhixm, epsij, self.f4, self.f5, self.f6)
    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    amono = xm * am

    # To be used in a1 and a2 for chain
    suma1_chain = suma1_monomer[:, diag_index[0], diag_index[1]]
    suma2_chain = suma2_monomer[:, diag_index[0], diag_index[1]]

    da1c = da1[:, :, diag_index[0], diag_index[1]]
    da2c = da2[:, :, diag_index[0], diag_index[1]]

    # Chain evaluation
    gHS = gdHS(x0i_matrix, xhix)
    gc = gammac(xhixm, self.alpha, tetha)

    # g1 sigma
    da1_chain = suma1_chain[1]
    suma_g1 = self.C * np.sum(da1c[0] * x0_g1, axis=0)
    g1s = g1sigma(xhi00, xm, da1_chain, suma_g1, a1vdw_cte)

    # g2 sigma
    dsum_a2new = suma2_chain
    da2new = da2new_dxhi00(khs, dkhs, dsum_a2new, self.eps)
    suma_g2 = self.C2 * np.sum(da2c[0] * x0_g2, axis=0)
    g2m = g2mca(xhi00, khs, xm, da2new, suma_g2, self.eps, a1vdw_cte)
    g2s = (1 + gc) * g2m

    lng = lngmie(gHS, g1s, g2s, beps, beps2)
    etai = xhi00 * xmi * di03[:, 3]
    achain = - np.dot(x * (self.ms - 1 + etai*self.ring), lng)

    ares = amono + achain

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        xj = x[self.compindex]
        Fab = temp_aux[25]
        aux_dii = temp_aux[26]
        aux_dii2 = temp_aux[27]
        Kab = temp_aux[28]
        iab = Iab(xhi, aux_dii, aux_dii2, Kab)
        Dab = self.sigmaij3 * Fab * iab
        Dabij = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab[self.indexab]

        Xass = Xass_solver(self.nsites, xj, rho, self.DIJ, Dabij,
                           self.diagasso, Xass)
        ares += np.dot(self.S * xj, (np.log(Xass) - Xass/2 + 1/2))
    else:
        Xass = None

    if self.polar_bool:
        eta = xhi[-1]
        epsa, epsija = temp_aux[29:]
        apolar = Apolar(rho, x, self.anij, self.bnij, self.cnij, eta,
                        epsa, epsija, self.sigma3, self.sigmaij3,
                        self.sigmaijk3,  self.npol, self.mupolad2)
        ares += apolar

    return ares, Xass


def dares_drho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3 = temp_aux[:10]
    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij = temp_aux[10:14]
    beps, beps2, a1vdw_cte, x0i_matrix, tetha = temp_aux[14:19]
    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii = temp_aux[19:25]

    dxhi00_drho = self.dxhi00_drho
    diag_index = self.diag_index

    xmi = x * self.ms
    xm = np.sum(xmi)
    # Eq A8
    xs = xmi / xm

    # Defining xhi0 wihtout x depedence
    xhi00 = dxhi00_drho * rho

    # Eq A7
    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs, xmi, xm, di03)
    # xhi x Eq A13
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs, xm, dij3)
    # xhi m Eq A23
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs, xm, self.sigmaij3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2., 6*xhix, 12*xhix2]])

    khs, dkhs, d2khs = d2kHS_dxhi00(xhix, dxhix_dxhi00)

    da1, da2 = d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij,
                                  J_lambdasij, self.cctesij, a1vdwij,
                                  a1vdw_cteij, dxhix_dxhi00)

    # Monomer evaluation
    suma1_monomer = self.Cij * np.sum(da1 * x0_a1, axis=1)
    suma2_monomer = self.Cij2 * np.sum(da2 * x0_a2, axis=1)

    a1ij = suma1_monomer[:2]
    a2ij = suma2_monomer[:2]
    aHS = dahs_dxhi00(xhi, dxhi_dxhi00)
    a1m = da1_dxhi00(xs, a1ij)
    a2m = da2_dxhi00(xs, khs, dkhs, xhixm, dxhixm_dxhi00, a2ij,
                     epsij, self.f1, self.f2, self.f3)
    a3m = da3_dxhi00(xs, xhixm, dxhixm_dxhi00, epsij, self.f4,
                     self.f5, self.f6)
    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    amono = xm * am

    # To be used in a1 and a2 of chain
    suma1_chain = suma1_monomer[:, diag_index[0], diag_index[1]]
    suma2_chain = suma2_monomer[:, diag_index[0], diag_index[1]]

    da1c = da1[:2, :, diag_index[0], diag_index[1]]
    da2c = da2[:2, :, diag_index[0], diag_index[1]]

    # chain evaluation
    gHS = dgdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)
    gc = dgammac_dxhi00(xhixm, dxhixm_dxhi00, self.alpha, tetha)

    # g1sigma
    da1_chain = suma1_chain[1:]
    dsuma_g1 = self.C * np.sum(da1c * x0_g1, axis=1)
    g1s = dg1sigma_dxhi00(xhi00, xm, da1_chain, dsuma_g1, a1vdw_cte)

    # g2sigma
    d2sum_a2new = suma2_chain
    d2a2new = d2a2new_dxhi00(khs, dkhs, d2khs, d2sum_a2new, self.eps)
    dsuma_g2 = self.C2 * np.sum(da2c * x0_g2, axis=1)
    g2m = dg2mca_dxhi00(xhi00, khs, dkhs, xm, d2a2new, dsuma_g2, self.eps,
                        a1vdw_cte)
    g2s = g2m * (1 + gc[0])
    g2s[1] += g2m[0] * gc[1]

    lng = dlngmie_dxhi00(gHS, g1s, g2s, beps, beps2)
    detai_dxhi00 = xmi * di03[:, 3]
    etai = xhi00 * detai_dxhi00
    achain = - lng@(x * (self.ms - 1. + etai*self.ring))
    achain[1] += - np.dot(x*self.ring*detai_dxhi00, lng[0])

    ares = amono + achain
    ares *= np.array([1, dxhi00_drho])

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        xj = x[self.compindex]
        Fab = temp_aux[25]
        aux_dii = temp_aux[26]
        aux_dii2 = temp_aux[27]
        Kab = temp_aux[28]
        iab, diab = dIab_drho(xhi, dxhi_dxhi00, dxhi00_drho, aux_dii,
                              aux_dii2, Kab)
        # Fab = np.exp(beta * self.eABij) - 1.
        Dab = self.sigmaij3 * Fab * iab
        dDab_drho = self.sigmaij3 * Fab * diab
        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab[self.indexab]
        dDabij_drho[self.indexabij] = dDab_drho[self.indexab]
        Xass = Xass_solver(self.nsites, xj, rho, self.DIJ, Dabij,
                           self.diagasso, Xass)
        CIJ = CIJ_matrix(rho, xj, Xass, self.DIJ, Dabij, self.diagasso)
        dXass = dXass_drho(rho, xj, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)
        ares[0] += np.dot(self.S * xj, (np.log(Xass) - Xass/2 + 1/2))
        ares[1] += np.dot(self.S*xj, (1/Xass - 1/2) * dXass)

    else:
        Xass = None

    if self.polar_bool:
        eta = xhi[-1]
        deta_dxhi00 = dxhi_dxhi00[-1]
        deta = deta_dxhi00 * self.dxhi00_drho
        epsa, epsija = temp_aux[29:]
        dapolar = dApolar_drho(rho, x, self.anij, self.bnij, self.cnij,
                               eta, deta, epsa, epsija,
                               self.sigma3, self.sigmaij3, self.sigmaijk3,
                               self.npol, self.mupolad2)
        ares += dapolar

    return ares, Xass


def d2ares_drho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3 = temp_aux[:10]
    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij = temp_aux[10:14]
    beps, beps2, a1vdw_cte, x0i_matrix, tetha = temp_aux[14:19]
    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii = temp_aux[19:25]

    dxhi00_drho = self.dxhi00_drho
    diag_index = self.diag_index

    xmi = x * self.ms
    xm = np.sum(xmi)
    # Equation A8
    xs = xmi / xm

    # Defining xhi0 without x depedence
    xhi00 = dxhi00_drho * rho

    # Equation A7
    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs, xmi, xm, di03)
    # xhi x Eq A13
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs, xm, dij3)
    # xhi m Eq A23
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs, xm, self.sigmaij3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3.*xhix2, 4.*xhix3],
                         [0., 2., 6*xhix, 12.*xhix2],
                         [0., 0., 6., 24.*xhix]])

    khs, dkhs, d2khs, d3khs = d3kHS_dxhi00(xhix, dxhix_dxhi00)

    da1, da2 = d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xm, I_lambdasij,
                                  J_lambdasij, self.cctesij, a1vdwij,
                                  a1vdw_cteij, dxhix_dxhi00)

    # Monomer evaluation
    suma1_monomer = self.Cij * np.sum(da1 * x0_a1, axis=1)
    suma2_monomer = self.Cij2 * np.sum(da2 * x0_a2, axis=1)

    suma_a1 = suma1_monomer[:3]
    suma_a2 = suma2_monomer[:3]

    aHS = d2ahs_dxhi00(xhi, dxhi_dxhi00)
    a1m = d2a1_dxhi00(xs,  suma_a1)
    a2m = d2a2_dxhi00(xs, khs, dkhs, d2khs, xhixm, dxhixm_dxhi00, suma_a2,
                      epsij, self.f1, self.f2, self.f3)
    a3m = d2a3_dxhi00(xs, xhixm, dxhixm_dxhi00, epsij, self.f4,
                      self.f5, self.f6)
    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    amono = xm * am

    # To be used in a1 and a2 of chain
    suma1_chain = suma1_monomer[:, diag_index[0], diag_index[1]]
    suma2_chain = suma2_monomer[:, diag_index[0], diag_index[1]]

    da1c = da1[:3, :, diag_index[0], diag_index[1]]
    da2c = da2[:3, :, diag_index[0], diag_index[1]]

    # Chain contribution
    gHS = d2gdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)
    gc = d2gammac_dxhi00(xhixm, dxhixm_dxhi00, self.alpha, tetha)

    # g1 sigma
    da1_chain = suma1_chain[1:]
    d2suma_g1 = self.C * np.sum(da1c * x0_g1, axis=1)
    g1s = d2g1sigma_dxhi00(xhi00, xm, da1_chain, d2suma_g1, a1vdw_cte)

    # g2 sigma
    d3a2new = d3a2new_dxhi00(khs, dkhs, d2khs, d3khs, suma2_chain, self.eps)
    d2suma_g2 = self.C2 * np.sum(da2c * x0_g2, axis=1)
    g2m = d2g2mca_dxhi00(xhi00, khs, dkhs, d2khs, xm, d3a2new, d2suma_g2,
                         self.eps, a1vdw_cte)
    g2s = g2m * (1. + gc[0])
    g2s[1] += g2m[0] * gc[1]
    g2s[2] += 2. * g2m[1] * gc[1] + g2m[0] * gc[2]

    lng = d2lngmie_dxhi00(gHS, g1s, g2s, beps, beps2)

    detai_dxhi00 = xmi * di03[:, 3]
    etai = xhi00 * detai_dxhi00
    achain = - lng@(x * (self.ms - 1. + etai*self.ring))
    aux_ring = x*self.ring*detai_dxhi00
    achain[1] += - np.dot(aux_ring, lng[0])
    achain[2] += - 2*np.dot(aux_ring, lng[1])

    ares = amono + achain

    ares *= np.array([1., dxhi00_drho, dxhi00_drho**2])

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        xj = x[self.compindex]
        Fab = temp_aux[25]
        aux_dii = temp_aux[26]
        aux_dii2 = temp_aux[27]
        Kab = temp_aux[28]
        iab, diab, d2iab = d2Iab_drho(xhi, dxhi_dxhi00, dxhi00_drho, aux_dii,
                                      aux_dii2, Kab)
        Dab = self.sigmaij3 * Fab * iab
        dDab_drho = self.sigmaij3 * Fab * diab
        d2Dab_drho = self.sigmaij3 * Fab * d2iab

        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])
        d2Dabij_drho = np.zeros([self.nsites, self.nsites])

        Dabij[self.indexabij] = Dab[self.indexab]
        dDabij_drho[self.indexabij] = dDab_drho[self.indexab]
        d2Dabij_drho[self.indexabij] = d2Dab_drho[self.indexab]

        Xass = Xass_solver(self.nsites, xj, rho, self.DIJ, Dabij,
                           self.diagasso, Xass)
        CIJ = CIJ_matrix(rho, xj, Xass, self.DIJ, Dabij, self.diagasso)
        dXass = dXass_drho(rho, xj, Xass, self.DIJ, Dabij, dDabij_drho,
                           CIJ)
        d2Xass = d2Xass_drho(rho, xj, Xass, dXass, self.DIJ, Dabij,
                             dDabij_drho, d2Dabij_drho, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        ares[0] += np.dot(self.S*xj, aux1)
        ares[1] += np.dot(self.S*xj, aux2 * dXass)
        ares[2] += np.dot(self.S*xj, -(dXass/Xass)**2+d2Xass*aux2)

    else:
        Xass = None

    if self.polar_bool:
        eta = xhi[-1]
        deta_dxhi00 = dxhi_dxhi00[-1]
        deta = deta_dxhi00 * self.dxhi00_drho
        epsa, epsija = temp_aux[29:]
        dapolar = d2Apolar_drho(rho, x, self.anij, self.bnij, self.cnij,
                                eta, deta, epsa, epsija,
                                self.sigma3, self.sigmaij3, self.sigmaijk3,
                                self.npol, self.mupolad2)
        ares += dapolar

    return ares, Xass


def dares_dx(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3 = temp_aux[:10]
    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij = temp_aux[10:14]
    beps, beps2, a1vdw_cte, x0i_matrix, tetha = temp_aux[14:19]
    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii = temp_aux[19:25]

    dxhi00_drho = self.dxhi00_drho
    diag_index = self.diag_index

    xmi = x * self.ms
    xm = np.sum(xmi)
    # Equation A8
    xs = xmi / xm
    dxs_dx = - np.multiply.outer(self.ms, self.ms * x) / xm**2
    dxs_dx[diag_index] += self.ms / xm

    # Defining xhi0 wihtout x dependence
    xhi00 = dxhi00_drho * rho

    # Equation A7
    out = dxhi_dx_eval(xhi00, xs, xmi, xm, self.ms, di03)
    xhi, dxhi_dxhi00, dxhi_dx = out
    # xhi x Eq A13
    out = dxhix_dx_eval(xhi00, xs, dxs_dx, xm, self.ms, dij3)
    xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00 = out
    # xhi m Eq A23
    out = dxhix_dx_eval(xhi00, xs, dxs_dx, xm, self.ms, self.sigmaij3)
    xhixm, dxhixm_dxhi00, dxhixm_dx, dxhixm_dx_dxhi00 = out

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2, 6*xhix, 12*xhix2]])

    khs, dkhs, dkhsx, dkhsxxhi = dkHS_dx_dxhi00(xhix, dxhix_dxhi00,
                                                dxhix_dx, dxhix_dx_dxhi00)

    out = da1sB_dx_dxhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xm, self.ms,
                                     I_lambdasij, J_lambdasij, self.cctesij,
                                     a1vdwij, a1vdw_cteij, dxhix_dxhi00,
                                     dxhix_dx, dxhix_dx_dxhi00)

    da1, da2, da1x, da2x, da1_xxhi00, da2_xxhi00 = out
    da1xxhi = da1_xxhi00[:, :, diag_index[0], diag_index[0]]
    da2xxhi = da2_xxhi00[:, :, diag_index[0], diag_index[0]]

    # monomer calculation
    suma1_monomer = self.Cij * np.sum(da1 * x0_a1, axis=1)
    suma2_monomer = self.Cij2 * np.sum(da2 * x0_a2, axis=1)

    suma1_monomerx = self.Cij * (da1x[0] * x0_a1[0] + da1x[1] * x0_a1[1])

    suma2_monomerx = da2x[0]*x0_a2[0] + da2x[1]*x0_a2[1]
    suma2_monomerx += da2x[2]*x0_a2[2]
    suma2_monomerx *= self.Cij2

    aHS, daHSx = dahs_dx(xhi, dxhi_dx)
    a1m, da1mx = da1_dx(xs, dxs_dx, suma1_monomer[0], suma1_monomerx)
    a2m, da2mx = da2_dx(xs, dxs_dx, khs, dkhsx, xhixm, dxhixm_dx,
                        suma2_monomer[0], suma2_monomerx, epsij,
                        self.f1, self.f2, self.f3)
    a3m, da3mx = da3_dx(xs, dxs_dx, xhixm, dxhixm_dx, epsij,
                        self.f4, self.f5, self.f6)

    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    damx = daHSx + beta * da1mx + beta2 * da2mx + beta3 * da3mx
    amono = xm * am
    damonox = self.ms * am + xm * damx

    # Chain contribution
    gHS, dgHSx = dgdHS_dx(x0i_matrix, xhix, dxhix_dx)
    gc, dgcx = dgammac_dx(xhixm, dxhixm_dx, self.alpha, tetha)

    suma1_chain = suma1_monomer[:, diag_index[0], diag_index[1]]
    suma2_chain = suma2_monomer[:, diag_index[0], diag_index[1]]

    da1c = da1[:, :, diag_index[0], diag_index[1]]
    da2c = da2[:, :, diag_index[0], diag_index[1]]

    # g1 sigma
    da1_chain = suma1_chain[1]
    da1x_chain = da1xxhi[0]*x0_a1ii[0]
    da1x_chain += da1xxhi[1]*x0_a1ii[1]
    da1x_chain *= self.C

    suma_g1 = self.C * np.sum(da1c[0] * x0_g1, axis=0)

    da1x_a = da1x[0]
    da1x_r = da1x[1]
    dg1x_a = da1x_a[:, diag_index[0], diag_index[1]]
    dg1x_r = da1x_r[:, diag_index[0], diag_index[1]]
    suma_g1x = dg1x_a*x0_g1[0] + dg1x_r*x0_g1[1]
    suma_g1x *= self.C

    g1s, dg1sx = dg1sigma_dx(xhi00, xm, self.ms, da1_chain, da1x_chain,
                             suma_g1, suma_g1x, a1vdw_cte)

    # g2 sigma
    suma_a2new = suma2_chain

    suma_a2x = suma2_monomerx[:, diag_index[0], diag_index[1]]

    suma_a2xxhi = da2xxhi[0] * x0_a2ii[0]
    suma_a2xxhi += da2xxhi[1] * x0_a2ii[1]
    suma_a2xxhi += da2xxhi[2] * x0_a2ii[2]
    suma_a2xxhi *= self.C2

    da2new, da2newx = da2new_dx_dxhi00(khs, dkhs, dkhsx, dkhsxxhi, suma_a2new,
                                       suma_a2x, suma_a2xxhi, self.eps)

    suma_g2 = self.C2 * np.sum(da2c[0] * x0_g2, axis=0)

    da2x_2a = da2x[0]
    da2x_2r = da2x[2]
    da2x_ar = da2x[1]
    dxa = da2x_2a[:, diag_index[0], diag_index[1]]
    dxar = da2x_ar[:, diag_index[0], diag_index[1]]
    dxr = da2x_2r[:, diag_index[0], diag_index[1]]
    suma_g2x = dxa * x0_g2[0]
    suma_g2x += dxar * x0_g2[1]
    suma_g2x += dxr * x0_g2[2]
    suma_g2x *= self.C2

    g2m, dg2mx = dg2mca_dx(xhi00, khs, dkhsx, xm, self.ms, da2new, da2newx,
                           suma_g2, suma_g2x, self.eps, a1vdw_cte)

    g2s = g2m * (1 + gc)
    dg2sx = dgcx * g2m + (1+gc)*dg2mx

    lng, dlngx = dlngmie_dx(gHS, g1s, g2s, dgHSx, dg1sx, dg2sx, beps, beps2)

    detai_dx = xhi00 * self.ms * di03[:, 3]
    etai = xhi00 * xmi * di03[:, 3]
    aux_ring = etai*self.ring
    aux_chain = x * (self.ms - 1 + aux_ring)
    achain = - lng@aux_chain
    dachainx = -  dlngx@aux_chain
    dachainx += - lng * (self.ms - 1. + aux_ring + self.ring*detai_dx*x)

    ares = amono + achain
    daresx = damonox + dachainx

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        xj = x[self.compindex]
        Fab = temp_aux[25]
        aux_dii = temp_aux[26]
        aux_dii2 = temp_aux[27]
        Kab = temp_aux[28]
        iab, diab = dIab_dx(xhi, dxhi_dx, aux_dii, aux_dii2, Kab)
        Dab = self.sigmaij3 * Fab * iab
        dDab_dx = self.sigmaij3 * Fab * diab
        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_dx = np.zeros([self.nc, self.nsites, self.nsites])

        Dabij[self.indexabij] = Dab[self.indexab]

        dDabij_dx[:, self.indexabij[0], self.indexabij[1]] = dDab_dx[:, self.indexab[0], self.indexab[1]]

        Xass = Xass_solver(self.nsites, xj, rho, self.DIJ, Dabij,
                           self.diagasso, Xass)
        CIJ = CIJ_matrix(rho, xj, Xass, self.DIJ, Dabij, self.diagasso)
        dXassx = dXass_dx(rho, xj, Xass, self.DIJ, Dabij, dDabij_dx,
                          self.dxjdx, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        aasso = np.dot(self.S*xj, aux1)
        daassox = (self.dxjdx * aux1 + dXassx * xj * aux2)@self.S
        ares += aasso
        daresx += daassox

    else:
        Xass = None
    if self.polar_bool:
        eta = xhi[-1]
        deta_dx = dxhi_dx[-1]
        epsa, epsija = temp_aux[29:]
        a, dax = dApolar_dx(rho, x, self.anij, self.bnij, self.cnij,
                            eta, deta_dx, epsa, epsija, self.sigma3,
                            self.sigmaij3, self.sigmaijk3,  self.npol,
                            self.mupolad2)
        ares += a
        daresx += dax

    return ares, daresx, Xass


def dares_dxrho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3 = temp_aux[:10]
    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij = temp_aux[10:14]
    beps, beps2, a1vdw_cte, x0i_matrix, tetha = temp_aux[14:19]
    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii = temp_aux[19:25]

    dxhi00_drho = self.dxhi00_drho
    diag_index = self.diag_index

    xmi = x * self.ms
    xm = np.sum(xmi)
    # Equation A8
    xs = xmi / xm
    dxs_dx = - np.multiply.outer(self.ms, self.ms * x) / xm**2
    dxs_dx[diag_index] += self.ms / xm

    # defining xhi0 without x dependence
    xhi00 = dxhi00_drho * rho

    # Equation A7
    xhi, dxhi_dxhi00, dxhi_dx = dxhi_dx_eval(xhi00, xs, xmi, xm,
                                             self.ms, di03)

    # xhi x Eq A13
    out = dxhix_dx_eval(xhi00, xs, dxs_dx, xm, self.ms, dij3)
    xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00 = out

    # xhi m Eq A23
    out = dxhix_dx_eval(xhi00, xs, dxs_dx, xm, self.ms, self.sigmaij3)
    xhixm, dxhixm_dxhi00, dxhixm_dx, dxhixm_dx_dxhi00 = out

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2, 6*xhix, 12*xhix2]])

    out = dkHS_dx_dxhi002(xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    khs, dkhs, d2khs, dkhsx, dkhsxxhi = out

    out = da1sB_dx_d2xhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xm, self.ms,
                                      I_lambdasij, J_lambdasij, self.cctesij,
                                      a1vdwij, a1vdw_cteij, dxhix_dxhi00,
                                      dxhix_dx, dxhix_dx_dxhi00)

    da1, da2, da1x, da2x, da1_xxhi00, da2_xxhi00 = out
    da1xxhi = da1_xxhi00[:, :, diag_index[0], diag_index[0]]
    da2xxhi = da2_xxhi00[:, :, diag_index[0], diag_index[0]]

    # monomer calculation
    suma1_monomer = self.Cij * np.sum(da1 * x0_a1, axis=1)
    suma2_monomer = self.Cij2 * np.sum(da2 * x0_a2, axis=1)

    suma1_monomerx = self.Cij * (da1x[0] * x0_a1[0] + da1x[1] * x0_a1[1])
    suma2_monomerx = da2x[0]*x0_a2[0] + da2x[1]*x0_a2[1]
    suma2_monomerx += da2x[2]*x0_a2[2]
    suma2_monomerx *= self.Cij2

    aHS, daHSx = dahs_dxxhi(xhi, dxhi_dxhi00, dxhi_dx)
    a1m, da1mx = da1_dxxhi(xs, dxs_dx, suma1_monomer[:2], suma1_monomerx)
    a2m, da2mx = da2_dxxhi(xs, dxs_dx, khs, dkhs, dkhsx, xhixm,
                           dxhixm_dxhi00, dxhixm_dx, suma2_monomer[:2],
                           suma2_monomerx, epsij,  self.f1, self.f2,
                           self.f3)
    a3m, da3mx = da3_dxxhi(xs, dxs_dx, xhixm, dxhixm_dxhi00, dxhixm_dx,
                           epsij, self.f4, self.f5, self.f6)

    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    damx = daHSx + beta * da1mx + beta2 * da2mx + beta3 * da3mx
    amono = xm * am
    damonox = self.ms * am[0] + xm * damx

    # Chain contribution
    gHS, dgHSx = dgdHS_dxxhi(x0i_matrix, xhix, dxhix_dxhi00, dxhix_dx)
    gc, dgcx = dgammac_dxxhi(xhixm, dxhixm_dxhi00, dxhixm_dx, self.alpha,
                             tetha)

    suma1_chain = suma1_monomer[:, diag_index[0], diag_index[1]]
    suma2_chain = suma2_monomer[:, diag_index[0], diag_index[1]]

    da1c = da1[:2, :, diag_index[0], diag_index[1]]
    da2c = da2[:2, :, diag_index[0], diag_index[1]]

    # g1 sigma
    da1_chain = suma1_chain[1:]
    da1x_chain = da1xxhi[0]*x0_a1ii[0]
    da1x_chain += da1xxhi[1]*x0_a1ii[1]
    da1x_chain *= self.C

    dsuma_g1 = self.C * np.sum(da1c * x0_g1, axis=1)

    da1x_a = da1x[0]
    da1x_r = da1x[1]
    dg1x_a = da1x_a[:, diag_index[0], diag_index[1]]
    dg1x_r = da1x_r[:, diag_index[0], diag_index[1]]
    suma_g1x = dg1x_a*x0_g1[0] + dg1x_r*x0_g1[1]
    suma_g1x *= self.C

    g1s, dg1sx = dg1sigma_dxxhi(xhi00, xm, self.ms, da1_chain, da1x_chain,
                                dsuma_g1, suma_g1x, a1vdw_cte)

    # g2 sigma
    suma_a2new = suma2_chain

    suma_a2x = suma2_monomerx[:, diag_index[0], diag_index[1]]

    suma_a2xxhi = da2xxhi[0] * x0_a2ii[0]
    suma_a2xxhi += da2xxhi[1] * x0_a2ii[1]
    suma_a2xxhi += da2xxhi[2] * x0_a2ii[2]
    suma_a2xxhi *= self.C2

    *d2a2new, da2newx = da2new_dxxhi_dxhi00(khs, dkhs, d2khs, dkhsx, dkhsxxhi,
                                            suma_a2new, suma_a2x, suma_a2xxhi,
                                            self.eps)

    dsuma_g2 = self.C2 * np.sum(da2c * x0_g2, axis=1)

    da2x_2a = da2x[0]
    da2x_2r = da2x[2]
    da2x_ar = da2x[1]
    dxa = da2x_2a[:, diag_index[0], diag_index[1]]
    dxar = da2x_ar[:, diag_index[0], diag_index[1]]
    dxr = da2x_2r[:, diag_index[0], diag_index[1]]
    suma_g2x = dxa * x0_g2[0]
    suma_g2x += dxar * x0_g2[1]
    suma_g2x += dxr * x0_g2[2]
    suma_g2x *= self.C2

    g2m, dg2mx = dg2mca_dxxhi(xhi00, khs, dkhs, dkhsx, xm, self.ms, d2a2new,
                              da2newx, dsuma_g2, suma_g2x, self.eps, a1vdw_cte)

    g2s = g2m * (1 + gc[0])
    g2s[1] += g2m[0] * gc[1]
    dg2sx = dgcx*g2m[0] + (1+gc[0])*dg2mx

    lng, dlngx = dlngmie_dxxhi(gHS, g1s, g2s, dgHSx, dg1sx, dg2sx, beps, beps2)

    detai_dxhi00 = xmi * di03[:, 3]
    etai = xhi00 * detai_dxhi00
    aux_ring = etai*self.ring
    aux_chain = x * (self.ms - 1 + aux_ring)

    achain = - lng@aux_chain
    achain[1] += - np.dot(x*self.ring*detai_dxhi00, lng[0])

    detai_dx = xhi00 * self.ms * di03[:, 3]
    dachainx = -  dlngx@aux_chain
    dachainx += - lng[0] * (self.ms - 1. + aux_ring + self.ring*detai_dx*x)

    ares = amono + achain
    daresx = damonox + dachainx

    ares *= np.array([1, dxhi00_drho])

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        xj = x[self.compindex]
        Fab = temp_aux[25]
        aux_dii = temp_aux[26]
        aux_dii2 = temp_aux[27]
        Kab = temp_aux[28]
        iab, diab, diabx = dIab_dxrho(xhi, dxhi_dxhi00, dxhi00_drho, dxhi_dx,
                                      aux_dii, aux_dii2, Kab)
        Dab = self.sigmaij3 * Fab * iab
        dDab_drho = self.sigmaij3 * Fab * diab
        dDab_dx = self.sigmaij3 * Fab * diabx

        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])
        dDabij_dx = np.zeros([self.nc, self.nsites, self.nsites])

        Dabij[self.indexabij] = Dab[self.indexab]
        dDabij_drho[self.indexabij] = dDab_drho[self.indexab]
        dDabij_dx[:, self.indexabij[0], self.indexabij[1]] = dDab_dx[:, self.indexab[0], self.indexab[1]]

        Xass = Xass_solver(self.nsites, xj, rho, self.DIJ, Dabij,
                           self.diagasso, Xass)
        CIJ = CIJ_matrix(rho, xj, Xass, self.DIJ, Dabij, self.diagasso)
        dXassx = dXass_dx(rho, xj, Xass, self.DIJ, Dabij, dDabij_dx,
                          self.dxjdx, CIJ)
        dXass = dXass_drho(rho, xj, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        aasso = np.dot(self.S*xj, aux1)
        daasso = np.dot(self.S*xj, aux2 * dXass)

        ares[0] += aasso
        ares[1] += daasso

        daassox = (self.dxjdx * aux1 + dXassx * xj * aux2)@self.S
        daresx += daassox
    else:
        Xass = None

    if self.polar_bool:
        eta = xhi[-1]
        deta_dxhi00 = dxhi_dxhi00[-1]
        deta = deta_dxhi00 * self.dxhi00_drho
        deta_dx = dxhi_dx[-1]
        epsa, epsija = temp_aux[29:]
        a, dax = dApolar_dxrho(rho, x, self.anij, self.bnij, self.cnij,
                               eta, deta, deta_dx, epsa, epsija,
                               self.sigma3, self.sigmaij3, self.sigmaijk3,
                               self.npol, self.mupolad2)
        ares += a
        daresx += dax

    return ares, daresx, Xass



from __future__ import division, print_function, absolute_import
import numpy as np
from ..math import gauss
from .tensionresult import TensionResult


def sgt_pure(rhov, rhol, Tsat, Psat, model, n=100, full_output=False,
             check_eq=True):
    """
    SGT for pure component (rhol, rhov, T, P) -> interfacial tension

    Parameters
    ----------
    rhov : float
        liquid phase density
    rhol : float
        vapor phase density
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    n : int, optional
        number of collocation points for IFT integration
    full_output : bool, optional
        wheter to outputs all calculation info
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria

    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    # roots and weights of Gauss quadrature
    roots, w = gauss(n)

    # cii = model.ci(Tsat)  # computing temperature dependent cii
    temp_aux = model.temperature_aux(Tsat)
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)

    rola = rhol * rofactor
    rova = rhov * rofactor
    # Tad = Tsat * Tfactor
    Pad = Psat * Pfactor

    # Equilibrium chemical potential
    mu0, Xass0 = model.muad_aux(rova, temp_aux)
    mu02, Xass1 = model.muad_aux(rola, temp_aux)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    roi = (rola-rova)*roots+rova
    wreal = np.abs((rola-rova)*w)

    dOm = np.zeros(n)
    i = 0
    dOm[i], Xass = model.dOm_aux(roi[i], temp_aux, mu0, Pad, Xass0)
    for i in range(1, n):
        dOm[i], Xass = model.dOm_aux(roi[i], temp_aux, mu0, Pad, Xass)

    tenint = np.nan_to_num(np.sqrt(2*dOm))
    ten = np.dot(wreal, tenint)
    ten *= tenfactor

    if full_output:
        zint = np.sqrt(1./(2*dOm))
        z = np.cumsum(wreal*zint)
        z /= zfactor
        roi /= rofactor
        dictresult = {'tension': ten, 'rho': roi, 'z': z,
                      'GPT': dOm}
        out = TensionResult(dictresult)
        return out

    return ten



from __future__ import division, print_function, absolute_import
import numpy as np
from ..equilibrium import bubblePy, lle, tpd_min, vlleb


def fobj_vle(model, Xexp, Yexp, Texp, Pexp, weights_vle=[1., 1.]):
    """
    Objective function to fit parameters for VLE in multicomponent mixtures
    """

    n = Pexp.shape[0]
    n1, n2 = Xexp.shape
    if n2 == n:
        Xexp = Xexp.T
        Yexp = Yexp.T

    P = np.zeros(n)  # n,
    Y = np.zeros_like(Xexp)  # nc,n

    for i in range(n):
        Y[i], P[i] = bubblePy(Yexp[i], Pexp[i], Xexp[i], Texp[i], model)

    error = weights_vle[0] * np.sum((Y-Yexp)**2)
    error += weights_vle[1] * np.sum((P/Pexp-1)**2)
    error /= n
    return error


def fobj_lle(model, Xexp, Wexp, Texp, Pexp, tpd=True, weights_lle=[1., 1.]):
    """
    Objective function to fit parameters for LLE in multicomponent mixtures
    """
    n = Pexp.shape[0]
    n1, n2 = Xexp.shape
    if n2 == n:
        Xexp = Xexp.T
        Wexp = Wexp.T

    X = np.zeros_like(Xexp)
    W = np.zeros_like(Wexp)
    Z = (Xexp+Wexp)/2

    for i in range(n):
        if tpd:
            X0, tpd = tpd_min(Xexp[i], Z[i], Texp[i], Pexp[i], model, 'L', 'L')
            W0, tpd = tpd_min(Wexp[i], Z[i], Texp[i], Pexp[i], model, 'L', 'L')
        else:
            X0 = Xexp[i]
            W0 = Wexp[i]
        X[i], W[i], beta = lle(X0, W0, Z[i], Texp[i], Pexp[i], model)

    error = weights_lle[0] * np.sum((X-Xexp)**2)
    error += weights_lle[1] * np.sum((W-Wexp)**2)
    error /= n
    return error


def fobj_vlleb(model, Xellv, Wellv, Yellv, Tellv, Pellv,
               weights_vlleb=[1., 1., 1., 1.]):
    """
    Objective function to fit parameters for VLLE in binary mixtures
    """
    n = len(Tellv)
    n1, n2 = Xellv.shape
    if n2 == n:
        Xellv = Xellv.T
        Wellv = Wellv.T
        Yellv = Yellv.T

    X = np.zeros_like(Xellv)
    W = np.zeros_like(Wellv)
    Y = np.zeros_like(Yellv)
    P = np.zeros(n)
    Zll = (Xellv + Wellv) / 2

    for i in range(n):
        try:
            X0, tpd = tpd_min(Xellv[i], Zll[i], Tellv[i], Pellv[i],
                              model, 'L', 'L')
            W0, tpd = tpd_min(Wellv[i], Zll[i], Tellv[i], Pellv[i],
                              model, 'L', 'L')
            X[i], W[i], Y[i], P[i] = vlleb(X0, W0, Yellv[i], Pellv[i],
                                           Tellv[i], 'T', model)
        except:
            pass

    error = weights_vlleb[0] * np.sum((X-Xellv)**2)
    error += weights_vlleb[1] * np.sum((W-Wellv)**2)
    error += weights_vlleb[2] * np.sum((Y-Yellv)**2)
    error += weights_vlleb[3] * np.sum((P/Pellv-1)**2)

    error /= n

    return error



from __future__ import division, print_function, absolute_import
import numpy as np


def lngmie(ghs, g1, g2, be, be2):
    # be = beta * eps
    lng = np.log(ghs) + (be*g1+be2*g2)/ghs
    return lng


def dlngmie_dxhi00(ghs, g1, g2, be, be2):
    # be = beta * eps
    ghs, dghs = ghs

    g1, dg1 = g1
    g2, dg2 = g2

    aux = (be*g1+be2*g2)

    lng = np.log(ghs) + aux/ghs

    dlng = be * ghs * (dg1 + be * dg2)
    dlng += dghs * (ghs - aux)
    dlng /= ghs**2

    return np.array([lng, dlng])


def d2lngmie_dxhi00(ghs, g1, g2, be, be2):
    # be = beta * eps

    ghs, dghs, d2ghs = ghs
    g1, dg1, d2g1 = g1
    g2, dg2, d2g2 = g2

    # be2 = be**2
    ghs2 = ghs**2
    ghs3 = ghs2 * ghs
    dghs2 = dghs**2
    aux1 = be*g1+be2*g2
    aux2 = be*dg1+be2*dg2

    lng = np.log(ghs) + aux1/ghs

    dlng = ghs * aux2
    dlng += dghs * (ghs - aux1)
    dlng /= ghs2
    '''
    d2lng = 2*be * dghs**2*(g1+be*g2)
    d2lng += ghs**2*(d2ghs + be * (d2g1+be*d2g2))
    d2lng += ghs * (-dghs*(2*be*(dg1+be*dg2)+dghs)-be*(g1+be*g2)*d2ghs)
    d2lng /= ghs**3
    '''
    d2lng = 2*dghs2*aux1
    d2lng += ghs2*(d2ghs + be * (d2g1+be*d2g2))
    d2lng += ghs*(-2*dghs*aux2-dghs2 - aux1*d2ghs)
    d2lng /= ghs3

    return np.array([lng, dlng, d2lng])


def dlngmie_dx(ghs, g1, g2, dghsx, dg1x, dg2x, be, be2):
    # be = beta * eps

    aux = be*g1+be2*g2
    lng = np.log(ghs) + aux/ghs

    dlng = be * ghs * (dg1x + be * dg2x)
    dlng += dghsx * (ghs - aux)
    dlng /= ghs**2

    return lng, dlng


def dlngmie_dxxhi(ghs, g1, g2, dghsx, dg1x, dg2x, be, be2):
    # be = beta * eps
    ghs, dghs = ghs

    g1, dg1 = g1
    g2, dg2 = g2
    aux = be*g1+be2*g2
    ghs2 = ghs**2

    lng = np.log(ghs) + aux / ghs

    dlng = be * ghs * (dg1 + be * dg2)
    dlng += dghs * (ghs - aux)
    dlng /= ghs2

    dlngx = be * ghs * (dg1x + be * dg2x)
    dlngx += dghsx * (ghs - aux)
    dlngx /= ghs2

    return np.array([lng, dlng]), dlngx



from __future__ import division, print_function, absolute_import


# Eq (20) Paper 2014
def B(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*Ikl/xhix13 - 9.*xhix*(1+xhix)*Jkl/(2.*xhix13)
    b *= -a1vdw_cte * xm * xhi00
    return b


def dB_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1

    b = (1-xhix/2)*Ikl/xhix_13 - 9.*xhix*(1+xhix)*Jkl/(2.*xhix_13)
    b *= -a1vdw_cte * xm * xhi00

    aux1 = xhix_1 * (-2*Ikl + xhix * (Ikl + 9*Jkl + 9*Jkl*xhix))
    aux2 = -5*Ikl+9*Jkl+xhix*(2*Ikl+36*Jkl+9*Jkl*xhix)

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= a1vdw_cte * xm
    db /= 2 * xhix_14
    return b, db


def d2B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1

    b = (1-xhix/2)*Ikl/xhix_13 - 9.*xhix*(1+xhix)*Jkl/(2.*xhix_13)
    b *= -a1vdw_cte * xm * xhi00

    aux1 = xhix_1 * (-2*Ikl + xhix * (Ikl + 9*Jkl + 9*Jkl*xhix))
    aux2 = -5*Ikl+9*Jkl+xhix*(2*Ikl+36*Jkl+9*Jkl*xhix)
    aux3 = aux2 * xhix_1
    aux4 = -3*(Ikl-4*Jkl)+xhix*(Ikl+21*Jkl+3*Jkl*xhix)

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= a1vdw_cte * xm
    db /= 2 * xhix_14

    d2b = aux3 * dxhix_dxhi00
    d2b += 3 * xhi00 * aux4 * dxhix_dxhi00**2
    d2b *= a1vdw_cte * xm
    d2b /= xhix_15
    return b, db, d2b


def d3B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1
    xhix_16 = xhix_15*xhix_1

    dxhix_dxhi00_2 = dxhix_dxhi00**2
    dxhix_dxhi00_3 = dxhix_dxhi00_2 * dxhix_dxhi00

    b = (1-xhix/2)*Ikl/xhix_13 - 9.*xhix*(1+xhix)*Jkl/(2.*xhix_13)
    b *= -a1vdw_cte * xm * xhi00

    aux1 = xhix_1 * (-2*Ikl + xhix * (Ikl + 9*Jkl + 9*Jkl*xhix))
    aux2 = -5*Ikl+9*Jkl+xhix*(2*Ikl+36*Jkl+9*Jkl*xhix)
    aux3 = aux2 * xhix_1
    aux4 = -3*(Ikl-4*Jkl)+xhix*(Ikl+21*Jkl+3*Jkl*xhix)
    aux5 = (-14*Ikl+81*Jkl+xhix*(4*Ikl+90*Jkl+9*Jkl*xhix))

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= a1vdw_cte * xm
    db /= 2 * xhix_14

    d2b = aux3 * dxhix_dxhi00
    d2b += 3 * xhi00 * aux4 * dxhix_dxhi00_2
    d2b *= a1vdw_cte * xm
    d2b /= xhix_15

    d3b = 9 * xhix_1 * aux4 * dxhix_dxhi00_2
    d3b += 3 * xhi00 * aux5 * dxhix_dxhi00_3
    d3b *= a1vdw_cte * xm
    d3b /= xhix_16

    return b, db, d2b, d3b



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from ..math import gdem
from .equilibriumresult import EquilibriumResult


# VLE phi-phi
def dew_sus(P_T, Y, T_P, dew_type, x_guess, eos, vl0, vv0,
            Xassl0, Xassv0, not_in_x=[]):

    if dew_type == 'T':
        P = P_T
        temp_aux = T_P
    elif dew_type == 'P':
        T = P_T
        temp_aux = eos.temperature_aux(T)
        P = T_P

    # Vapour fugacities
    lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V', vv0, Xassv0)
    if np.any(np.isnan(lnphiv)) or np.isnan(vv):
        lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V')

    tol = 1e-8
    error = 1
    itacc = 0
    niter = 0
    n = 5
    X_calc = x_guess
    X = x_guess

    # Liquid fugacities
    lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L', vl0, Xassl0)
    if np.any(np.isnan(lnphil)) or np.isnan(vl):
        lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L')

    while error > tol and itacc < 3:
        niter += 1

        lnK = lnphil-lnphiv
        K = np.exp(lnK)
        X_calc_old = X_calc
        X_calc = Y/K

        # modification for non-condensable components Ki -> infty
        X_calc[not_in_x] = 0.

        if niter == (n-3):
            X3 = X_calc
        elif niter == (n-2):
            X2 = X_calc
        elif niter == (n-1):
            X1 = X_calc
        elif niter == n:
            niter = 0
            itacc += 1
            dacc = gdem(X_calc, X1, X2, X3)
            X_calc += dacc
        error = np.linalg.norm(X_calc - X_calc_old)
        X = X_calc/X_calc.sum()
        # Liquid fugacitiies
        lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L', vl, Xassl)
        if np.any(np.isnan(lnphil)) or np.isnan(vl):
            lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L')

    if dew_type == 'T':
        f0 = X_calc.sum() - 1
    elif dew_type == 'P':
        f0 = np.log(X_calc.sum())

    return f0, X, lnK, vl, vv, Xassl, Xassv


def dew_newton(inc, Y, T_P, dew_type, eos, in_x=[]):

    global vl, vv, Xassl, Xassv

    f = np.zeros_like(inc)
    lnK = inc[:-1]
    K = np.exp(lnK)

    if dew_type == 'T':
        P = inc[-1]
        temp_aux = T_P
    elif dew_type == 'P':
        T = inc[-1]
        temp_aux = eos.temperature_aux(T)
        P = T_P

    nc = eos.nc
    X = np.zeros(nc)
    X[in_x] = Y[in_x] / K

    # Liquid fugacities
    lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L', vl, Xassl)
    if np.any(np.isnan(lnphil)) or np.isnan(vl):
        lnphil, vl, Xassl = eos.logfugef_aux(X, temp_aux, P, 'L')
    # Vapor fugacities
    lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V', vv, Xassv)
    if np.any(np.isnan(lnphiv)) or np.isnan(vv):
        lnphiv, vv, Xassv = eos.logfugef_aux(Y, temp_aux, P, 'V')

    f[:-1] = lnK + (lnphiv - lnphil)[in_x]
    f[-1] = (Y-X).sum()

    return f


def dewPx(x_guess, P_guess, y, T, model, good_initial=False,
          v0=[None, None], Xass0=[None, None],
          not_in_x_list=[],
          full_output=False):
    """
    Dew point (T, y) -> (P, y)

    Solves dew point at given vapour composition and temperature. It uses a
    combination of accelerated successive sustitution with quasi Newton Method
    in regular cases and when good initial it's provided the full system of
    equations of the phase envelope method is used as objective function.

    Parameters
    ----------
    x_guess : array_like
        guess of liquid phase composition
    P_guess : float
        guess of equilibrium pressure of the liquid [Pa]
    y : array_like
        vapour phase composition
    T : float
        temperaure of the vapour [K]
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip succesive substitution and solves by Newton's Method.
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    Xass0 : list, optional
        if supplied X association non-bonded fraction sites initial value
    not_in_x_list : list, optional
        index of components not present in the liquid phase (default is empty)
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        liquid mole fraction vector
    P : float
        equilibrium pressure [Pa]

    """
    nc = model.nc
    if len(x_guess) != nc or len(y) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    if np.any([i > (nc-1) for i in not_in_x_list]):
        raise Exception('Index of components not_in_y_list must be less than nc')

    not_in_x = np.zeros(nc, dtype=bool)
    not_in_x[not_in_x_list] = 1
    in_x = np.logical_not(not_in_x)

    temp_aux = model.temperature_aux(T)

    global vl, vv, Xassl, Xassv
    vl0, vv0 = v0
    Xassl0, Xassv0 = Xass0

    it = 0
    itmax = 10
    tol = 1e-8

    P = P_guess
    out = dew_sus(P, y, temp_aux, 'T', x_guess, model, vl0, vv0,
                  Xassl0, Xassv0, not_in_x=not_in_x)
    f, X, lnK, vl, vv, Xassl, Xassv = out
    error = np.abs(f)
    h = 1e-3

    method = 'quasi-newton + ASS'
    while error > tol and it <= itmax and not good_initial:
        it += 1
        out = dew_sus(P+h, y, temp_aux, 'T', X, model, vl, vv, Xassl, Xassv,
                      not_in_x=not_in_x)
        f1, X1, lnK1, vl, vv, Xassl, Xassv = out
        out = dew_sus(P, y, temp_aux, 'T', X, model, vl, vv, Xassl, Xassv, 
                      not_in_x=not_in_x)
        f, X, lnK, vl, vv, Xassl, Xassv = out
        df = (f1-f)/h
        dP = f / df
        if dP > P:
            dP = 0.4 * P
        elif np.isnan(dP):
            dP = 0.0
            it = 1.*itmax
        P -= dP
        error = np.abs(f)

    if error > tol:
        inc0 = np.hstack([lnK[in_x], P])
        sol1 = root(dew_newton, inc0, args=(y, temp_aux, 'T', model, in_x))
        if sol1.success:
            method = 'second-order'
            error = np.linalg.norm(sol1.fun)
            it += sol1.nfev

            sol = sol1.x
            lnK_newton = sol[:-1]
            K_newton = np.exp(lnK_newton)
            X = np.zeros(nc)
            X[in_x] = y[in_x] / K_newton
            X /= np.sum(X)
            P = sol[-1]

            rhol, Xassl = model.density_aux(X, temp_aux, P, 'L', rho0=1./vl,
                                            Xass0=Xassl)
            rhov, Xassv = model.density_aux(y, temp_aux, P, 'V', rho0=1./vv,
                                            Xass0=Xassv)
            vl = 1./rhol
            vv = 1./rhov

    if full_output:
        sol = {'T': T, 'P': P, 'error': error, 'iter': it,
               'X': X, 'v1': vl, 'Xassl': Xassl, 'state1': 'Liquid',
               'Y': y, 'v2': vv, 'Xassv': Xassv, 'state2': 'Vapor',
               'method': method}
        out = EquilibriumResult(sol)
        return out

    return X, P


def dewTx(x_guess, T_guess, y, P, model, good_initial=False,
          v0=[None, None], Xass0=[None, None],
          not_in_x_list=[],
          full_output=False):
    """
    Dew point (T, y) -> (P, y)

    Solves dew point at given vapour composition and pressure. It uses a
    combination of accelerated successive sustitution with quasi Newton Method
    in regular cases and when good initial it's provided the full system of
    equations of the phase envelope method is used as objective function.

    Parameters
    ----------
    x_guess : array_like
        guess of liquid phase composition
    T_guess : float
        guess of equilibrium temperature of the liquid [K]
    y : array_like
        vapour phase composition
    P : float
        pressure of the liquid [Pa]
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip succesive substitution and solves by Newton's Method.
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    Xass0 : list, optional
        if supplied X association non-bonded fraction sites initial value
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        liquid mole fraction vector
    T : float
        equilibrium temperature [K]

    """

    nc = model.nc
    if len(x_guess) != nc or len(y) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    if np.any([i > (nc-1) for i in not_in_x_list]):
        raise Exception('Index of components not_in_y_list must be less than nc')

    not_in_x = np.zeros(nc, dtype=bool)
    not_in_x[not_in_x_list] = 1
    in_x = np.logical_not(not_in_x)

    global vl, vv, Xassl, Xassv
    vl0, vv0 = v0
    Xassl0, Xassv0 = Xass0

    it = 0
    itmax = 10
    tol = 1e-8

    T = T_guess
    out = dew_sus(T, y, P, 'P', x_guess, model, vl0, vv0, 
                  Xassl0, Xassv0, not_in_x=not_in_x)
    f, X, lnK, vl, vv, Xassl, Xassv = out
    error = np.abs(f)
    h = 1e-4
    method = 'quasi-newton + ASS'
    while error > tol and it <= itmax and not good_initial:
        it += 1
        out = dew_sus(T+h, y, P, 'P', X, model, vl, vv, 
                      Xassl, Xassv, not_in_x=not_in_x)
        f1, X1, lnK1, vl, vv, Xassl, Xassv = out
        out = dew_sus(T, y, P, 'P', X, model, vl, vv, 
                      Xassl, Xassv, not_in_x=not_in_x)
        f, X, lnK, vl, vv, Xassl, Xassv = out
        df = (f1-f)/h
        if np.isnan(df):
            df = 0.0
            it = 1.*itmax
        T -= f/df
        error = np.abs(f)

    if error > tol:
        inc0 = np.hstack([lnK[in_x], T])
        sol1 = root(dew_newton, inc0, args=(y, P, 'P', model, in_x))
        if sol1.success:
            method = 'second-order'
            error = np.linalg.norm(sol1.fun)
            it += sol1.nfev

            sol = sol1.x
            lnK_newton = sol[:-1]
            K_newton = np.exp(lnK_newton)
            X = np.zeros(nc)
            X[in_x] = y[in_x] / K_newton
            X /= np.sum(X)
            T = sol[-1]

            temp_aux = model.temperature_aux(T)
            rhol, Xassl = model.density_aux(X, temp_aux, P, 'L', rho0=1./vl,
                                            Xass0=Xassl)
            rhov, Xassv = model.density_aux(y, temp_aux, P, 'V', rho0=1./vv,
                                            Xass0=Xassv)
            vl = 1./rhol
            vv = 1./rhov

    if full_output:
        sol = {'T': T, 'P': P, 'error': error, 'iter': it,
               'X': X, 'v1': vl, 'Xassl': Xassl, 'state1': 'Liquid',
               'Y': y, 'v2': vv, 'Xassv': Xassv, 'state2': 'Vapor',
               'method': method}
        out = EquilibriumResult(sol)
        return out

    return X, T


__all__ = ['dewTx', 'dewPx']


from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import fsolve
from ..math import lobatto, colocA
from .cijmix_cy import cmix_cy
from .tensionresult import TensionResult


def fobj_beta0(ro, ro_s, s, temp_aux, mu0, sqrtci, model):
    nc = model.nc
    ro = np.insert(ro, s, ro_s)
    global Xass
    mu, Xass = model.muad_aux(ro, temp_aux, Xass)
    dmu = mu - mu0

    f1 = sqrtci[s]*dmu
    f2 = sqrtci*dmu[s]

    return (f1-f2)[np.arange(nc) != s]


def ten_beta0_reference(rho1, rho2, Tsat, Psat, model, s=0,
                        n=100, full_output=False, check_eq=True):

    nc = model.nc

    # Dimensionless profile
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    ro1a = rho1*rofactor
    ro2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]
    ci = np.diag(cij)
    sqrtci = np.sqrt(ci)

    temp_aux = model.temperature_aux(Tsat)

    mu0, Xass01 = model.muad_aux(ro1a, temp_aux)

    mu02, Xass02 = model.muad_aux(ro2a, temp_aux)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    # roots and weights for Lobatto quadrature
    roots, weights = lobatto(n)

    ro_s = (ro2a[s]-ro1a[s])*roots+ro1a[s]  # Integration nodes
    wreal = np.abs(weights*(ro2a[s]-ro1a[s]))  # Integration weights

    # A matrix for derivatives with orthogonal collocation
    A = colocA(roots) / (ro2a[s]-ro1a[s])

    rodep = np.zeros([nc-1, n])
    rodep[:, 0] = ro1a[np.arange(nc) != s]

    global Xass
    Xass = Xass01
    for i in range(1, n):
        rodep[:, i] = fsolve(fobj_beta0, rodep[:, i-1],
                             args=(ro_s[i], s, temp_aux, mu0, sqrtci, model))
    ro = np.insert(rodep, s, ro_s, axis=0)
    dro = rodep@A.T
    dro = np.insert(dro, s, np.ones(n), axis=0)

    suma = cmix_cy(dro, cij)

    dom = np.zeros(n)
    Xass = Xass01
    for k in range(1, n-1):
        dom[k], Xass = model.dOm_aux(ro[:, k], temp_aux, mu0, Pad, Xass)

    intten = np.nan_to_num(np.sqrt(suma*(2*dom)))
    ten = np.dot(intten, wreal)
    ten *= tenfactor

    if full_output:
        # Z profile
        with np.errstate(divide='ignore'):
            intz = (np.sqrt(suma/(2*dom)))
        intz[np.isinf(intz)] = 0
        z = np.cumsum(intz*wreal)
        z /= zfactor
        ro /= rofactor
        dictresult = {'tension': ten, 'rho': ro, 'z': z,
                      'GPT': dom}
        out = TensionResult(dictresult)
        return out

    return ten



from __future__ import division, print_function, absolute_import
import numpy as np


# Expansion of polar term
aij = np.array([[0.3043504, 0.9534641, -1.1610080],
               [-0.1358588, -1.8396383, 4.5258607],
               [1.4493329, 2.0131180, 0.9751222],
               [0.3556977, -7.3724958, -12.281038],
               [-2.0653308, 8.2374135, 5.9397575]])

bij = np.array([[0.2187939, -0.5873164, 3.4869576],
               [-1.1896431, 1.2489132, -14.915974],
               [1.1626889, -0.508528, 15.372022],
               [0., 0., 0.],
               [0., 0., 0.]])

cij = np.array([[-0.0646774, -0.9520876, -0.6260979],
               [0.1975882, 2.9924258, 1.2924686],
               [-0.8087562, -2.3802636, 1.652783],
               [0.6902849, -0.2701261, -3.4396744],
               [0.,  0., 0.]])


def Apolar(rho, x, anij, bnij, cnij,
           eta, epsa, epsija, sigma3, sigmaij3, sigmaijk3,  npol, muad2):

    etavec = np.power(eta, [0, 1, 2, 3, 4])

    J2DDij = np.tensordot(anij + bnij * epsija, etavec, axes=((0), (0)))

    ter1 = x * sigma3 * epsa * npol * muad2
    outer_ter1 = np.outer(ter1, ter1)
    ter2 = - np.pi * outer_ter1 / sigmaij3
    A2 = rho * np.sum(ter2 * J2DDij)

    J3DDijk = np.tensordot(cnij, etavec, axes=((0), (0)))

    ter3 = np.multiply.outer(ter1, outer_ter1) / sigmaijk3

    A3 = rho**2 * np.sum(ter3 * J3DDijk)
    A3 *= -4 * np.pi**2 / 3

    with np.errstate(divide='ignore', invalid='ignore'):
        Apol = np.nan_to_num(A2/(1 - A3/A2))

    return Apol


def dApolar_drho(rho, x, anij, bnij, cnij, eta, deta, epsa, epsija, sigma3,
                 sigmaij3, sigmaijk3,  npol, muad2):

    etavec = np.array([[1., eta, eta**2, eta**3, eta**4],
                      [0., 1., 2*eta, 3*eta**2, 4*eta**3]])

    J2DDij, dJ2DDij = np.tensordot(etavec, anij+bnij*epsija, axes=((1), (0)))
    dJ2DDij *= deta

    J3DDijk, dJ3DDijk = np.tensordot(etavec, cnij, axes=((1), (0)))
    dJ3DDijk *= deta

    ter1 = x * sigma3 * epsa * npol * muad2
    outer_ter1 = np.outer(ter1, ter1)
    ter2 = - np.pi * outer_ter1 / sigmaij3
    A2 = rho * np.sum(ter2 * J2DDij)
    dA2 = np.sum(ter2 * (J2DDij + rho * dJ2DDij))

    ter2 = np.multiply.outer(ter1, outer_ter1) / sigmaijk3

    A3 = rho**2 * np.sum(ter2 * J3DDijk)
    A3 *= -4 * np.pi**2 / 3
    dA3 = np.sum(ter2 * rho * (2 * J3DDijk + rho * dJ3DDijk))
    dA3 *= -4 * np.pi**2 / 3

    with np.errstate(divide='ignore', invalid='ignore'):
        Apol = np.nan_to_num(A2/(1 - A3/A2))
        dApolar = A2 * ((A2 - 2 * A3) * dA2 + A2 * dA3)
        dApolar /= ((A2 - A3)**2)

    A = np.nan_to_num(np.array([Apol, dApolar]))

    return A


def d2Apolar_drho(rho, x, anij, bnij, cnij, eta, deta, epsa, epsija, sigma3,
                  sigmaij3, sigmaijk3,  npol, muad2):

    etavec = np.array([[1., eta, eta**2, eta**3, eta**4],
                      [0., 1., 2*eta, 3*eta**2, 4*eta**3],
                      [0., 0., 2., 6*eta, 12*eta**2]])

    J2DDij, dJ2DDij, d2J2DDij = np.tensordot(etavec, anij+bnij*epsija,
                                             axes=((1), (0)))
    dJ2DDij *= deta
    d2J2DDij *= deta**2

    J3DDijk, dJ3DDijk, d2J3DDijk = np.tensordot(etavec, cnij, axes=((1), (0)))
    dJ3DDijk *= deta
    d2J3DDijk *= deta**2

    ter1 = x * sigma3 * epsa * npol * muad2
    outer_ter1 = np.outer(ter1, ter1)
    ter2 = - np.pi * outer_ter1 / sigmaij3
    A2 = rho * np.sum(ter2 * J2DDij)
    dA2 = np.sum(ter2 * (J2DDij + rho * dJ2DDij))
    d2A2 = np.sum(ter2 * (2 * dJ2DDij + rho * d2J2DDij))

    ter2 = np.multiply.outer(ter1, outer_ter1) / sigmaijk3

    A3 = rho**2 * np.sum(ter2 * J3DDijk)
    A3 *= -4 * np.pi**2 / 3
    dA3 = np.sum(ter2 * rho * (2 * J3DDijk + rho * dJ3DDijk))
    dA3 *= -4 * np.pi**2 / 3
    d2A3 = np.sum(ter2 * (2 * J3DDijk + rho * (4*dJ3DDijk+rho*d2J3DDijk)))
    d2A3 *= -4 * np.pi**2 / 3

    with np.errstate(divide='ignore', invalid='ignore'):
        Apol = (A2/(1 - A3/A2))

        dApolar = A2 * ((A2 - 2 * A3) * dA2 + A2 * dA3)
        dApolar /= ((A2 - A3)**2)

        d2Apolar = 2 * A3**2 * (dA2**2 + A2 * d2A2)
        d2Apolar += A2**2 * (2 * dA3**2 + A2 * (d2A2 + d2A3))
        d2Apolar -= A2 * A3 * (4 * dA2 * dA3 + A2 * (3*d2A2 + d2A3))
        d2Apolar /= ((A2 - A3)**3)

    A = np.nan_to_num(np.array([Apol, dApolar, d2Apolar]))

    return A


def dApolar_dx(rho, x, anij, bnij, cnij, eta, deta_dx, epsa, epsija, sigma3,
               sigmaij3, sigmaijk3,  npol, muad2):

    etavec = np.array([[1., eta, eta**2, eta**3, eta**4],
                      [0., 1., 2*eta, 3*eta**2, 4*eta**3]])

    J2DDij, dJ2DDij = np.tensordot(etavec, anij+bnij*epsija, axes=((1), (0)))
    dJ2DDij_dx = np.multiply.outer(deta_dx, dJ2DDij)

    J3DDijk, dJ3DDijk = np.tensordot(etavec, cnij, axes=((1), (0)))
    dJ3DDijk_dx = np.multiply.outer(deta_dx, dJ3DDijk)

    ter1 = x * sigma3 * epsa * npol * muad2
    outer_ter1 = np.outer(ter1, ter1)
    ter2 = - np.pi * outer_ter1 / sigmaij3
    A2 = rho * np.sum(ter2 * J2DDij)

    ter1x = sigma3 * epsa * npol * muad2
    outer_ter1x = np.outer(ter1x, ter1x)
    ter2x = - np.pi * outer_ter1x * J2DDij/sigmaij3
    dA2x = rho * (2*ter2x@x + np.sum(ter2 * dJ2DDij_dx, axis=(1, 2)))

    ter2 = np.multiply.outer(ter1, outer_ter1) / sigmaijk3

    A3 = rho**2 * np.sum(ter2 * J3DDijk)
    A3 *= -4 * np.pi**2 / 3

    ter2x = np.multiply.outer(ter1x, outer_ter1x) / sigmaijk3
    dA3x = 3*x@(ter2x * J3DDijk)@x + np.sum(ter2 * dJ3DDijk_dx, axis=(1, 2, 3))
    dA3x *= -4 * np.pi**2 / 3 * rho**2

    with np.errstate(divide='ignore', invalid='ignore'):
        Apol = np.nan_to_num(A2/(1 - A3/A2))
        dApolarx = A2 * ((A2 - 2 * A3) * dA2x + A2 * dA3x)
        dApolarx /= ((A2 - A3)**2)

    return Apol, np.nan_to_num(dApolarx)


def dApolar_dxrho(rho, x, anij, bnij, cnij, eta, deta, deta_dx, epsa, epsija,
                  sigma3, sigmaij3, sigmaijk3,  npol, muad2):

    etavec = np.array([[1., eta, eta**2, eta**3, eta**4],
                      [0., 1., 2*eta, 3*eta**2, 4*eta**3]])

    J2DDij, dJ2DDij = np.tensordot(etavec, anij+bnij*epsija, axes=((1), (0)))
    dJ2DDij_dx = np.multiply.outer(deta_dx, dJ2DDij)
    dJ2DDij *= deta

    J3DDijk, dJ3DDijk = np.tensordot(etavec, cnij, axes=((1), (0)))
    dJ3DDijk_dx = np.multiply.outer(deta_dx, dJ3DDijk)
    dJ3DDijk *= deta

    ter1 = x * sigma3 * epsa * npol * muad2
    outer_ter1 = np.outer(ter1, ter1)
    ter2 = - np.pi * outer_ter1 / sigmaij3
    A2 = rho * np.sum(ter2 * J2DDij)
    dA2 = np.sum(ter2 * (J2DDij + rho * dJ2DDij))

    ter1x = sigma3 * epsa * npol * muad2
    outer_ter1x = np.outer(ter1x, ter1x)
    ter2x = - np.pi * outer_ter1x * J2DDij/sigmaij3
    dA2x = rho * (2*ter2x@x + np.sum(ter2 * dJ2DDij_dx, axis=(1, 2)))

    ter2 = np.multiply.outer(ter1, outer_ter1) / sigmaijk3

    A3 = rho**2 * np.sum(ter2 * J3DDijk)
    A3 *= -4 * np.pi**2 / 3

    dA3 = np.sum(ter2 * rho * (2 * J3DDijk + rho * dJ3DDijk))
    dA3 *= -4 * np.pi**2 / 3

    ter2x = np.multiply.outer(ter1x, outer_ter1x) / sigmaijk3
    dA3x = 3*x@(ter2x * J3DDijk)@x + np.sum(ter2 * dJ3DDijk_dx, axis=(1, 2, 3))
    dA3x *= -4 * np.pi**2 / 3 * rho**2

    with np.errstate(divide='ignore', invalid='ignore'):
        Apol = A2/(1 - A3/A2)

        dApolar = A2 * ((A2 - 2 * A3) * dA2 + A2 * dA3)
        dApolar /= ((A2 - A3)**2)

        dApolarx = A2 * ((A2 - 2 * A3) * dA2x + A2 * dA3x)
        dApolarx /= ((A2 - A3)**2)

    A = np.nan_to_num(np.array([Apol, dApolar]))

    return A, np.nan_to_num(dApolarx)



from __future__ import division, print_function, absolute_import
import numpy as np

# Equation A12
def B(xhi00, xhix, xm, I_ij, J_ij, a1vdw_cteij):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix13 = (1-xhix)**3
    b = (1-xhix/2)*I_ij/xhix13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix13)
    b *= -a1vdw_cteij * xm * xhi00
    return b


def dB_dxhi00(xhi00, xhix, xm, I_ij, J_ij, a1vdw_cteij, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)
    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1

    aux0 = a1vdw_cteij * xm
    aux1 = xhix_1 * (-2*I_ij + xhix * (I_ij + 9*J_ij + 9*J_ij*xhix))
    aux2 = -5*I_ij+9*J_ij+xhix*(2*I_ij+36*J_ij+9*J_ij*xhix)

    b = (1-xhix/2)*I_ij/xhix_13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix_13)
    b *= - aux0 * xhi00

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= aux0
    db /= 2 * xhix_14

    return b, db


def d2B_dxhi00(xhi00, xhix, xm, I_ij, J_ij, a1vdw_cteij, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1

    aux0 = a1vdw_cteij * xm
    aux1 = xhix_1 * (-2*I_ij + xhix * (I_ij + 9*J_ij + 9*J_ij*xhix))
    aux2 = -5*I_ij+9*J_ij+xhix*(2*I_ij+36*J_ij+9*J_ij*xhix)
    aux3 = xhix_1 * aux2
    aux4 = -3*(I_ij-4*J_ij)+xhix*(I_ij+21*J_ij+3*J_ij*xhix)

    b = (1-xhix/2)*I_ij/xhix_13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix_13)
    b *= - aux0 * xhi00

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= aux0
    db /= 2 * xhix_14

    d2b = aux3 * dxhix_dxhi00
    d2b += 3 * xhi00 * aux4 * dxhix_dxhi00**2
    d2b *= aux0
    d2b /= xhix_15

    return b, db, d2b


def d3B_dxhi00(xhi00, xhix, xm, I_ij, J_ij, a1vdw_cteij, dxhix_dxhi00):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1
    xhix_16 = xhix_15*xhix_1

    dxhix_dxhi00_2 = dxhix_dxhi00**2
    dxhix_dxhi00_3 = dxhix_dxhi00_2 * dxhix_dxhi00

    aux0 = a1vdw_cteij * xm
    aux1 = xhix_1 * (-2*I_ij + xhix * (I_ij + 9*J_ij + 9*J_ij*xhix))
    aux2 = -5*I_ij+9*J_ij+xhix*(2*I_ij+36*J_ij+9*J_ij*xhix)
    aux3 = aux2 * xhix_1
    aux4 = -3*(I_ij-4*J_ij)+xhix*(I_ij+21*J_ij+3*J_ij*xhix)
    aux5 = (-14*I_ij+81*J_ij+xhix*(4*I_ij+90*J_ij+9*J_ij*xhix))

    b = (1-xhix/2)*I_ij/xhix_13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix_13)
    b *= -aux0 * xhi00

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= aux0
    db /= 2 * xhix_14

    d2b = aux3 * dxhix_dxhi00
    d2b += 3 * xhi00 * aux4 * dxhix_dxhi00_2
    d2b *= aux0
    d2b /= xhix_15

    d3b = 9 * xhix_1 * aux4 * dxhix_dxhi00_2
    d3b += 3 * xhi00 * aux5 * dxhix_dxhi00_3
    d3b *= aux0
    d3b /= xhix_16

    return b, db, d2b, d3b


def dB_dx_dxhi00_dxxhi(xhi00, xhix, xm, ms, I_ij, J_ij, a1vdw_cte,
                       dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1

    b = (1-xhix/2)*I_ij/xhix_13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix_13)
    b *= -a1vdw_cte * xm * xhi00

    aux1 = xhix_1 * (-2*I_ij + xhix * (I_ij + 9*J_ij + 9*J_ij*xhix))
    aux2 = -5*I_ij+9*J_ij+xhix*(2*I_ij+36*J_ij+9*J_ij*xhix)
    # aux3 = aux2 * xhix_1

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= a1vdw_cte * xm
    db /= 2 * xhix_14

    bx = np.multiply.outer(ms, aux1)
    bx += np.multiply.outer(dxhix_dx, aux2*xm)
    bx *= a1vdw_cte * xhi00
    bx /= 2 * xhix_14

    aux4 = aux1 + xhi00 * aux2 * dxhix_dxhi00
    aux4 *= xhix_1

    aux5 = (-3*(I_ij-4*J_ij)+xhix*(I_ij+21*J_ij+3*J_ij*xhix))
    aux5 *= 6*xhi00*dxhix_dxhi00
    aux5 += aux2 * xhix_1
    aux5 *= xm

    aux6 = aux2 * xhix_1 * xm * xhi00

    bxxhi = np.multiply.outer(ms, aux4)
    bxxhi += np.multiply.outer(dxhix_dx, aux5)
    bxxhi += np.multiply.outer(dxhix_dx_dxhi00, aux6)
    bxxhi *= a1vdw_cte
    bxxhi /= (2 * xhix_15)
    return b, db, bx, bxxhi


def dB_dx_d2xhi00_dxxhi(xhi00, xhix, xm, ms, I_ij, J_ij, a1vdw_cte,
                        dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00):

    xhix_1 = 1-xhix
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1

    b = (1-xhix/2)*I_ij/xhix_13 - 9.*xhix*(1+xhix)*J_ij/(2.*xhix_13)
    b *= -a1vdw_cte * xm * xhi00

    aux1 = xhix_1 * (-2*I_ij + xhix * (I_ij + 9*J_ij + 9*J_ij*xhix))
    aux2 = -5*I_ij+9*J_ij+xhix*(2*I_ij+36*J_ij+9*J_ij*xhix)
    aux3 = aux2 * xhix_1
    aux4 = -3*(I_ij-4*J_ij)+xhix*(I_ij+21*J_ij+3*J_ij*xhix)

    db = aux1 + xhi00 * dxhix_dxhi00 * aux2
    db *= a1vdw_cte * xm
    db /= 2 * xhix_14

    bx = np.multiply.outer(ms, aux1)
    bx += np.multiply.outer(dxhix_dx, aux2*xm)
    bx *= a1vdw_cte * xhi00
    bx /= 2 * xhix_14

    aux5 = aux1 + xhi00 * aux2 * dxhix_dxhi00
    aux5 *= xhix_1

    aux6 = aux2 * xhix_1
    aux6 += 6*xhi00*aux4*dxhix_dxhi00
    aux6 *= xm

    aux7 = aux2 * xhix_1 * xm * xhi00

    bxxhi = np.multiply.outer(ms, aux5)
    bxxhi += np.multiply.outer(dxhix_dx, aux6)
    bxxhi += np.multiply.outer(dxhix_dx_dxhi00, aux7)
    bxxhi *= a1vdw_cte
    bxxhi /= (2 * xhix_15)

    d2b = aux3 * dxhix_dxhi00
    d2b += 3 * xhi00 * aux4 * dxhix_dxhi00**2
    d2b *= a1vdw_cte * xm
    d2b /= xhix_15
    return b, db, d2b, bx, bxxhi



from __future__ import division, print_function, absolute_import
import numpy as np


# Hard sphere Eq A6
def ahs(xhi):
    xhi0, xhi1, xhi2, xhi3 = xhi
    xhi3_1 = 1-xhi3
    xhi23 = xhi2**3

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_1**2
    a /= xhi0
    return a


def dahs_dxhi00(xhi, dxhi_dxhi00):
    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    xhi23 = xhi2**3

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    da = xhi2 * dxhi2**2 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * -xhi3_1
    da -= dxhi0 * dxhi3 * xhi3_12
    da /= -xhi3_13 * dxhi0
    return np.array([a, da])


def d2ahs_dxhi00(xhi, dxhi_dxhi00):

    xhi0, xhi1, xhi2, xhi3 = xhi
    dxhi0, dxhi1, dxhi2, dxhi3 = dxhi_dxhi00
    xhi3_1 = 1-xhi3
    xhi3_12 = xhi3_1**2
    xhi3_13 = xhi3_12*xhi3_1
    xhi3_14 = xhi3_13*xhi3_1
    xhi23 = xhi2**3
    dxhi22 = dxhi2**2
    dxhi23 = dxhi22*dxhi2

    a = (xhi23/xhi3**2 - xhi0) * np.log(xhi3_1)
    a += 3 * xhi1 * xhi2 / xhi3_1
    a += xhi23 / xhi3 / xhi3_12
    a /= xhi0

    da = xhi2 * dxhi22 * (-3 + xhi3)
    da += 3 * dxhi1 * dxhi2 * -xhi3_1
    da -= dxhi0 * dxhi3 * xhi3_12
    da /= -xhi3_13 * dxhi0

    d2a = -6 * dxhi1 * dxhi2 * dxhi3 * -xhi3_1
    d2a += dxhi0 * dxhi3**2 * xhi3_12
    d2a += dxhi23 * (3 + xhi3 * (4 - xhi3))
    d2a /= xhi3_14 * dxhi0
    return np.array([a, da, d2a])



from __future__ import division, print_function, absolute_import
import numpy as np


def da2new_dxhi00(khs, dkhs, da2, eps):

    ctes = eps / 2
    sum1, dsum1 = da2
    da2 = sum1 * dkhs + khs * dsum1
    da2 *= ctes

    return da2


def d2a2new_dxhi00(khs, dkhs, d2khs, d2a2, eps):

    ctes = eps / 2
    sum1, dsum1, d2sum1 = d2a2

    da2 = sum1 * dkhs + khs * dsum1
    da2 *= ctes

    d2a2 = sum1 * d2khs + khs * d2sum1 + 2 * dsum1 * dkhs
    d2a2 *= ctes

    return da2, d2a2


def d3a2new_dxhi00(khs, dkhs, d2khs, d3khs, d3a2, eps):

    ctes = eps/2
    sum1, dsum1, d2sum1, d3sum1 = d3a2

    da2 = sum1 * dkhs + khs * dsum1
    da2 *= ctes

    d2a2 = sum1 * d2khs + khs * d2sum1 + 2 * dsum1 * dkhs
    d2a2 *= ctes

    d3a2 = 3 * (dsum1 * d2khs + dkhs * d2sum1)
    d3a2 += sum1 * d3khs + khs * d3sum1
    d3a2 *= ctes

    return da2, d2a2, d3a2



from __future__ import division, print_function, absolute_import
import numpy as np
from .path_sk import ten_beta0_sk
from .reference_component import ten_beta0_reference


def sgt_mix_beta0(rho1, rho2, Tsat, Psat, model, n=100, method='reference',
                  full_output=False, s=0, alpha0=None, check_eq=True):
    """
    SGT for mixtures and beta = 0 (rho1, rho2, T, P) -> interfacial tension

    Parameters
    ----------
    rho1 : float
        phase 1 density vector
    rho2 : float
        phase 2 density vector
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    n : int, optional
        number points to solve density profiles
    method : string
        method used to solve density profiles, available options are
        'reference' for using reference component method and 'liang'
        for Liang path function
    full_output : bool, optional
        wheter to outputs all calculation info
    s : int
        index of reference component used in refernce component method
    alpha0 : float
        initial guess for solving Liang path function
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria


    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    cij = model.ci(Tsat)
    cij /= cij[0, 0]
    dcij = np.linalg.det(cij)
    if not np.isclose(dcij, 0):
        warning = 'Influece parameter matrix is not singular probably a'
        warning += ' beta has been set up.'
        raise Exception(warning)

    if method == 'reference':
        sol = ten_beta0_reference(rho1, rho2, Tsat, Psat, model,
                                  s, n, full_output, check_eq)
    elif method == 'liang':
        sol = ten_beta0_sk(rho1, rho2, Tsat, Psat, model, n, full_output,
                           alpha0, check_eq)
    else:
        raise Warning("method not known")

    return sol



from __future__ import division, print_function, absolute_import
import numpy as np
from ..math import gdem
from scipy.optimize import minimize
from .equilibriumresult import EquilibriumResult
from ..constants import Na


def rachfordrice(beta, K, Z, tol=1e-8, maxiter=20, not_in_x=[], not_in_y=[]):
    '''
    Solves Rachford Rice equation by Halley's method
    '''
    K1 = K-1.

    ZK = Z*K
    ZK[not_in_y] = 0.0
    ZK[not_in_x] = 1e4  # setting it to a high number

    Z_div_K = Z/K
    Z_div_K[not_in_y] = 1e4  # setting it to a high number
    Z_div_K[not_in_x] = 0.0

    g0 = np.sum(ZK) - 1.
    g1 = 1. - np.sum(Z_div_K)

    singlephase = False

    if g0 < 0:
        beta = 0.
        D = np.ones_like(Z)
        singlephase = True
    elif g1 > 0:
        beta = 1.
        D = 1 + K1
        singlephase = True
    it = 0
    e = 1.

    while e > tol and it < maxiter and not singlephase:
        it += 1
        D = 1. + beta*K1
        KD = K1/D

        # modification for not_in_y components Ki -> 0
        KD[not_in_y] = -1. / (1. - beta)

        # modification for not_in_y components Ki -> infty
        KD[not_in_x] = 1. / beta

        fo = np.dot(Z, KD)
        dfo = - np.dot(Z, KD**2)
        d2fo = 2*np.dot(Z, KD**3)
        dbeta = - (2*fo*dfo)/(2*dfo**2-fo*d2fo)
        beta += dbeta
        e = np.abs(dbeta)

    return beta, D, singlephase


def Gibbs_obj(ny_var, phases, Z, nx, ny, where_x, where_y,
              where_equilibria, temp_aux, P, model):
    '''
    Objective function to minimize Gibbs energy in biphasic flash
    '''
    ny[where_equilibria] = ny_var
    nx[where_equilibria] = Z[where_equilibria] - ny_var

    X = nx / np.sum(nx)
    Y = ny / np.sum(ny)

    global vx, vy, Xass_x, Xass_y
    with np.errstate(all='ignore'):
        lnfugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, phases[0], vx,
                                                Xass_x)
        lnfugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, phases[1], vy,
                                                Xass_y)
        # If the previous volume/Xass initial guess is not good it might
        # cause the whole calculation to fail
        if np.any(np.isnan(lnfugx)) or np.isnan(vx):
            lnfugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, phases[0])
        if np.any(np.isnan(lnfugy)) or np.isnan(vy):
            lnfugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, phases[1])

        fugx = np.log(X) + lnfugx
        fugy = np.log(Y) + lnfugy

    fx = np.dot(nx[where_x], fugx[where_x])
    fy = np.dot(ny[where_y], fugy[where_y])

    f = fx + fy
    with np.errstate(all='ignore'):
        df = (fugy - fugx)[where_equilibria]
    return f, df


def Qfun_obj_uc(inc, nx, ny, where_equilibria, nc_equilibria, Z, temp_aux, Pspec, model):
    '''
    Objective function to minimize Q function = A + Pspec * V in biphasic flash
    '''
    RT = Na / temp_aux[0]
    Pspec_by_RT = Pspec / RT

    ny_var = inc[:nc_equilibria]
    nx_var = Z[where_equilibria] - ny_var

    ny[where_equilibria] = ny_var
    nx[where_equilibria] = nx_var

    lnvy_total = inc[nc_equilibria]
    lnvx_total = inc[nc_equilibria+1]

    vy_total = np.exp(lnvy_total)
    vx_total = np.exp(lnvx_total)

    ny_total = np.sum(ny)
    nx_total = np.sum(nx)
    vy = vy_total / ny_total
    vx = vx_total / nx_total

    global Xass_x, Xass_y
    with np.errstate(all='ignore'):
        #### phase Y
        y = ny / ny_total
        rho_y = 1./vy
        rhom_y = Na * rho_y
        afcn_y, Xass_y = model.afcn_aux(y, rhom_y, temp_aux, Xass_y)
        afcn_y /= RT
        afcn_y *= ny_total

        #### phase X
        x = nx / nx_total
        rho_x = 1./vx
        rhom_x = Na * rho_x
        afcn_x, Xass_x = model.afcn_aux(x, rhom_x, temp_aux, Xass_x)
        afcn_x /= RT
        afcn_x *= nx_total

    ## Objective function
    A_total = afcn_y + afcn_x
    PV_total = (vy_total + vx_total) * Pspec_by_RT
    Q_total = A_total + PV_total
    return Q_total


def dQfun_obj_uc(inc, nx, ny, where_equilibria, nc_equilibria, Z, temp_aux, Pspec, model):
    '''
    Objective function to minimize Q function = A + Pspec * V in biphasic flash
    '''
    RT = Na / temp_aux[0]
    Pspec_by_RT = Pspec / RT

    ny_var = inc[:nc_equilibria]
    nx_var = Z[where_equilibria] - ny_var

    ny[where_equilibria] = ny_var
    nx[where_equilibria] = nx_var

    lnvy_total = inc[nc_equilibria]
    lnvx_total = inc[nc_equilibria+1]

    vy_total = np.exp(lnvy_total)
    vx_total = np.exp(lnvx_total)

    ny_total = np.sum(ny)
    nx_total = np.sum(nx)
    vy = vy_total / ny_total
    vx = vx_total / nx_total

    global Xass_x, Xass_y
    with np.errstate(all='ignore'):
        #### phase Y
        y = ny / ny_total
        rho_y = 1./vy
        rhom_y = Na * rho_y
        a_y, ax_y, Xass_y = model.dafcn_dxrho_aux(y, rhom_y, temp_aux, Xass_y)
        ax_y = np.nan_to_num(ax_y)
        a_y /= RT
        ax_y /= RT
        afcn_y, dafcn_y = a_y

        Zy = dafcn_y * rhom_y
        Py_by_RT = Zy / vy
        mu_nvt_y = afcn_y + ax_y - np.dot(y, ax_y) + Zy
        afcn_y *= ny_total

        #### phase X
        x = nx / nx_total
        rho_x = 1./vx
        rhom_x = Na * rho_x
        a_x, ax_x, Xass_x = model.dafcn_dxrho_aux(x, rhom_x, temp_aux, Xass_x)
        ax_x = np.nan_to_num(ax_x)
        a_x /= RT
        ax_x /= RT
        afcn_x, dafcn_x = a_x

        Zx = dafcn_x * rhom_x
        Px_by_RT = Zx / vx
        mu_nvt_x = afcn_x + ax_x - np.dot(x, ax_x) + Zx
        afcn_x *= nx_total

    ## Objective function
    A_total = afcn_y + afcn_x
    PV_total = (vy_total + vx_total) * Pspec_by_RT
    Q_total = A_total + PV_total

    # gradient
    dQ = np.zeros_like(inc)
    dQ[:nc_equilibria] = mu_nvt_y[where_equilibria] - mu_nvt_x[where_equilibria]
    dQ[nc_equilibria] = vy_total * (- Py_by_RT + Pspec_by_RT)
    dQ[nc_equilibria+1] = vx_total * (- Px_by_RT + Pspec_by_RT)

    return Q_total, dQ


def flash(x_guess, y_guess, equilibrium, Z, T, P, model, v0=[None, None],
          Xass0=[None, None], K_tol=1e-8, nacc=3, accelerate_every=5,
          not_in_x_list=[], not_in_y_list=[],
          maxiter_radfordrice=10, tol_rachfordrice=1e-8,
          good_initial=False,
          minimization_approach='gibbs',
          minimization_method='BFGS',
          minimization_tol=1e-12,
          minimization_options={},
          full_output=False):
    """
    Isothermic isobaric flash (z, T, P) -> (x,y,beta)

    Parameters
    ----------

    x_guess : array_like
        guess composition of phase 1
    y_guess : array_like
        guess composition of phase 2
    equilibrium : string
        'LL' for LLE, 'LV' for VLE
    z : array_like
        overall system composition
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= Y/X) vector
    nacc : int, optional
        number of accelerated successive substitution cycles to perform
    accelerate_every : int, optional
        number of iterations to perform before acceleration in successive substitution
        must be equal or greater than 4
    not_in_x_list : list, optional
        index of components not present in phase 1 (default is empty)
    not_in_y_list : list, optional
        index of components not present in phase 2 (default is empty)
    maxiter_radfordrice : int, optional
        maximum number of iterations for Rachford-Rice equation
    tol_rachfordrice : float, optional
        tolerance for Rachford-Rice solver
    good_initial : bool, optional
        if True, the function assumes that the initial guesses are good and proceeds
        to minimize the Gibbs energy without performing the accelerated successive substitution
    minimization_approach : string, optional
        'gibbs' or 'helmholtz' minimization approach
    minimization_method : string, optional
        minimization method to use in the minimization of the Gibbs energy 
        only used if minimization_approach='gibbs'
        see scipy.minimize for possible minimizers
    minimization_tol : float, optional
        tolerance for the minimization of the Gibbs energy
    minimization_options : dict, optional
        options for scipy.minimize function
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------
    X : array_like
        phase 1 composition
    Y : array_like
        phase 2 composition
    beta : float
        phase 2 phase fraction
    """

    nc = model.nc
    if len(x_guess) != nc or len(y_guess) != nc or len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    if np.any([i > (nc-1) for i in not_in_y_list]):
        raise Exception('Index of components not_in_y_list must be less than nc')

    if np.any([i > (nc-1) for i in not_in_y_list]):
        raise Exception('Index of components not_in_x_list must be less than nc')

    # creating list for non-condensable/non-volatiles
    not_in_x = np.zeros(nc, dtype=bool)
    not_in_x[not_in_x_list] = 1
    in_x = np.logical_not(not_in_x)

    not_in_y = np.zeros(nc, dtype=bool)
    not_in_y[not_in_y_list] = 1
    in_y = np.logical_not(not_in_y)

    in_equilibria = np.logical_and(in_x, in_y)

    # setting up temperature and auxiliary variables
    temp_aux = model.temperature_aux(T)
    # setting up initial guesses
    X = x_guess
    Y = y_guess
    vx0, vy0 = v0
    Xass_x0, Xass_y0 = Xass0

    # setting up initial values for volume roots
    global vx, vy, Xass_x, Xass_y

    with np.errstate(all='ignore'):
        fugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, equilibrium[0],
                                              vx0, Xass_x0)
        fugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, equilibrium[1],
                                              vy0, Xass_y0)

        # If the previous volume/Xass initial guess is not good it might
        # cause the whole calculation to fail
        if np.any(np.isnan(fugx)) or np.isnan(vx):
            fugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, equilibrium[0])
        if np.any(np.isnan(fugy)) or np.isnan(vy):
            fugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, equilibrium[1])

    lnK = fugx - fugy
    K = np.exp(lnK)

    bmin = max(np.hstack([((K*Z-1.)/(K-1.))[K > 1], Z[not_in_x], 0.]))
    bmax = min(np.hstack([((1.-Z)/(1.-K))[K < 1], 1.]))
    beta_new = (bmin + bmax)/2

    # Even if good guesses are given it is good to do one Radford-Rice iteration
    beta, D, singlephase = rachfordrice(beta_new, K, Z,
                                        tol=tol_rachfordrice,
                                        maxiter=maxiter_radfordrice,
                                        not_in_y=not_in_y,
                                        not_in_x=not_in_x)
    # updating phase compositions
    X = Z/D
    Y = X*K
    # modification for for non-condensable/non-volatiles
    if not singlephase:
        # modification for not_in_y components Ki -> 0
        X[not_in_y] = Z[not_in_y] / (1. - beta)
        # modification for not_in_x components Ki -> infty
        Y[not_in_x] = Z[not_in_x] / beta
    Y[not_in_y] = 0.
    X[not_in_x] = 0.
    X /= X.sum()
    Y /= Y.sum()

    ##########################################
    # 1. Accelerated Successive Substitution #
    ##########################################
    method = "ASS"

    # dummy variables to track iterarions and errors
    e1 = 100.
    it_acc = 0
    it_total = 1
    it_max = nacc * accelerate_every

    while e1 > K_tol and it_total <= it_max and not good_initial:
        it_total += 1
        it_acc += 1
        lnK_old = 1. * lnK

        # updating fugacity coefficients
        with np.errstate(all='ignore'):
            fugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, equilibrium[0],
                                                  vx, Xass_x)
            fugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, equilibrium[1],
                                                  vy, Xass_y)
            # If the previous volume/Xass initial guess is not good it might
            # cause the whole calculation to fail
            if np.any(np.isnan(fugx)) or np.isnan(vx):
                fugx, vx, Xass_x = model.logfugef_aux(X, temp_aux, P, equilibrium[0])
            if np.any(np.isnan(fugy)) or np.isnan(vy):
                fugy, vy, Xass_y = model.logfugef_aux(Y, temp_aux, P, equilibrium[1])
        lnK = fugx - fugy

        # acceleration
        if it_acc == (accelerate_every-3):
            lnK3 = lnK
        elif it_acc == (accelerate_every-2):
            lnK2 = lnK
        elif it_acc == (accelerate_every-1):
            lnK1 = lnK
        elif it_acc == accelerate_every:
            dacc = gdem(lnK, lnK1, lnK2, lnK3)
            lnK += dacc
            # reset the acceleration counter
            it_acc = 0

        K = np.exp(lnK)
        e1 = np.sum((lnK-lnK_old)**2, where=in_equilibria)

        # getting a feasible beta (if not in the feasible range)
        bmin = max(np.hstack([((K*Z-1.)/(K-1.))[K > 1], Z[not_in_x], 0.]))
        bmax = min(np.hstack([((1.-Z)/(1.-K))[K < 1], 1.]))
        if beta < bmin or beta > bmax:
            beta_new = (bmin + bmax) / 2.

        # solving Rachford-Rice equation
        beta, D, singlephase = rachfordrice(beta_new, K, Z,
                                            tol=tol_rachfordrice,
                                            maxiter=maxiter_radfordrice,
                                            not_in_y=not_in_y,
                                            not_in_x=not_in_x)

        # updating phase compositions
        X = Z/D
        Y = X*K
        # modification for for non-condensable/non-volatiles
        if not singlephase:
            # modification for not_in_y components Ki -> 0
            X[not_in_y] = Z[not_in_y] / (1. - beta)
            # modification for not_in_x components Ki -> infty
            Y[not_in_x] = Z[not_in_x] / beta
        Y[not_in_y] = 0.
        X[not_in_x] = 0.
        X /= X.sum()
        Y /= Y.sum()

    ###################################
    # 2. Minimization of Gibbs energy #
    ###################################

    # if e1 > K_tol and it_total == it_max and not singlephase:
    if e1 > K_tol and not singlephase:
        # in case the minimization does not converge, we save the previous values
        vx_copy = 1. * vx
        vy_copy = 1. * vy
        Xass_x_copy = 1. * Xass_x
        Xass_y_copy = 1. * Xass_y

        # If a componnent is not in the global phase, it is not considered in the minimization  
        z_notzero = Z != 0

        where_x = np.logical_and(z_notzero, in_x)
        where_y = np.logical_and(z_notzero, in_y)
        where_equilibria = np.logical_and(z_notzero, in_equilibria)

        # setting up initial guesses
        ny = np.zeros(nc)
        nx = np.zeros(nc)

        ny[not_in_y] = 0.
        ny[not_in_x] = Z[not_in_x]

        nx[not_in_x] = 0.
        nx[not_in_y] = Z[not_in_y]

        ny_var = beta * Y[where_equilibria]

        # PT flash by minimizing Gibbs energy
        if minimization_approach == 'gibbs':
            fobj = Gibbs_obj
            jac = True
            hess = None

            # bounds = [(0., ny_max) for ny_max in Z[where_equilibria]]

            args = (equilibrium, Z, nx, ny, where_x, where_y, where_equilibria, temp_aux, P, model)
            ny_sol = minimize(fobj, ny_var, jac=jac, method=minimization_method, hess=hess,
                              tol=minimization_tol, args=args, options=minimization_options)
            if ny_sol.success or np.mean(ny_sol.jac**2) < minimization_tol:
                method = "Gibbs_minimization"
                ny_var = ny_sol.x
                ny[where_equilibria] = ny_var
                nx[where_equilibria] = Z[where_equilibria] - ny_var
                beta = np.sum(ny)
                X = nx / np.sum(nx)
                Y = ny / beta

                # updating volume roots for founded equilibria compositions and volumes
                rho_x, Xass_x = model.density_aux(X, temp_aux, P, equilibrium[0],
                                                  rho0=1./vx, Xass0=Xass_x)
                rho_y, Xass_y = model.density_aux(Y, temp_aux, P, equilibrium[1],
                                                  rho0=1./vy, Xass0=Xass_y)
                vx = 1./rho_x
                vy = 1./rho_y

                it_total += ny_sol.nit
                e1 = np.linalg.norm(ny_sol.jac)
            else:
                vx = vx_copy
                vy = vy_copy
                Xass_x = Xass_x_copy
                Xass_y = Xass_y_copy

        #  PT flash by minimizing Gibbs energy (using a helmholtz approach)
        elif minimization_approach == 'helmholtz':
            # setting up initial guesses
            nc_equilibria = np.sum(where_equilibria)
            nx_var = Z[where_equilibria] - ny_var

            ny[where_equilibria] = ny_var
            nx[where_equilibria] = nx_var

            ny_total = np.sum(ny)
            nx_total = np.sum(nx)
            vy_total = vy*ny_total
            vx_total = vx*nx_total

            lnvy_total = np.log(vy_total)
            lnvx_total = np.log(vx_total)

            inc = np.hstack([ny_var, lnvy_total, lnvx_total])

            bounds = [(0, Z[i]) for i in range(nc) if where_equilibria[i]]
            bounds += [(None, None), (None, None)]

            args = (nx, ny, where_equilibria, nc_equilibria, Z, temp_aux, P, model)

            # If the initial guess is "poor" the jacobian might suggest a wrong direction
            # This is just to improve the initial guess without any derivative information
            Qsol_init = minimize(Qfun_obj_uc, inc, args=args, jac=False,
                                 method='SLSQP', bounds=bounds, options={'maxiter': 15})
            it_total += Qsol_init.nit
            Qsol = Qsol_init
            jac = Qsol_init.jac

            # Now that we have a better initial guess, we can use the jacobian
            # TNC method seems to work well after this step but the final solution might not be super accurate
            Qsol_tnc = minimize(dQfun_obj_uc, Qsol_init.x, args=args,
                                jac=True, method='TNC', bounds=bounds,
                                options=minimization_options)
            if Qsol_tnc.fun < Qsol.fun:
                it_total += Qsol_tnc.nit
                Qsol = Qsol_tnc
                jac = Qsol_tnc.jac
            # Refinining the solution. SLSQP works well when there is a very good initial guess
            if np.mean(jac**2) > minimization_tol:
                Qsol_slsqp = minimize(dQfun_obj_uc, Qsol_tnc.x, args=args,
                                      jac=True, method='SLSQP',
                                      bounds=bounds, tol=minimization_tol)
                if Qsol_slsqp.fun < Qsol.fun:
                    it_total += Qsol_slsqp.nit
                    Qsol = Qsol_slsqp

            if Qsol.success or np.mean(Qsol.jac**2) < minimization_tol:
                method = "Helmholtz_minimization"
                inc = Qsol.x

                ny_var = inc[:nc_equilibria]
                nx_var = Z[where_equilibria] - ny_var

                ny[where_equilibria] = ny_var
                nx[where_equilibria] = nx_var

                lnvy_total = inc[nc_equilibria]
                lnvx_total = inc[nc_equilibria+1]

                vy_total = np.exp(lnvy_total)
                vx_total = np.exp(lnvx_total)

                ny_total = np.sum(ny)
                nx_total = np.sum(nx)
                vy = vy_total / ny_total
                vx = vx_total / nx_total

                beta = ny_total
                Y = ny / ny_total
                X = nx / nx_total

                # updating volume roots for the obtained equilibria compositions and volumes
                rho_x, Xass_x = model.density_aux(X, temp_aux, P, equilibrium[0],
                                                  rho0=1./vx, Xass0=Xass_x)
                rho_y, Xass_y = model.density_aux(Y, temp_aux, P, equilibrium[1],
                                                  rho0=1./vy, Xass0=Xass_y)
                vx = 1./rho_x
                vy = 1./rho_y

                jac = Qsol.jac
                e1 = np.linalg.norm(jac)
                it_total += Qsol.nit
            else:
                vx = vx_copy
                vy = vy_copy
                Xass_x = Xass_x_copy
                Xass_y = Xass_y_copy

        else:
            raise Exception('Minimization approach not recognized. Use "gibbs" or "helmholtz"')

    if beta == 1.0:
        X = Y.copy()
    elif beta == 0.:
        Y = X.copy()

    if full_output:
        sol = {'T': T, 'P': P, 'beta': beta, 'error': e1, 'iter': it_total,
               'X': X, 'v1': vx, 'Xass1': Xass_x, 'state1': equilibrium[0],
               'Y': Y, 'v2': vy, 'Xass2': Xass_y, 'state2': equilibrium[1],
               'method': method}
        out = EquilibriumResult(sol)
        return out

    return X, Y, beta



from __future__ import division, print_function, absolute_import
import numpy as np
from .B_monomer import B, dB, d2B, d3B
from .a1s_monomer import a1s, da1s, d2a1s, d3a1s


def a1sB(eta, Ilam, Jlam, lam, cctes, eps):
    b = B(eta, Ilam, Jlam, eps)
    a1 = a1s(eta, lam, cctes, eps)
    a1b = a1 + b
    return a1b


def da1sB(eta, Ilam, Jlam, lam, cctes, eps):
    db = dB(eta, Ilam, Jlam, eps)
    da1 = da1s(eta, lam, cctes, eps)

    da1b = da1 + db

    return da1b


def d2a1sB(eta, Ilam, Jlam, lam, cctes, eps):
    d2b = d2B(eta, Ilam, Jlam, eps)
    d2a1 = d2a1s(eta, lam, cctes, eps)
    d2a1b = d2a1 + d2b
    return d2a1b


def d3a1sB(eta, Ilam, Jlam, lam, cctes, eps):
    d3b = d3B(eta, Ilam, Jlam, eps)
    d3a1 = d3a1s(eta, lam, cctes, eps)
    d3a1b = d3a1 + d3b
    return d3a1b


# Function for creating arrays with a1s + B at all lambdas
def a1B_eval(eta, I_lambdas, J_lambdas, lambdas, cctes, eps):
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdas
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdas
    lambda_a, lambda_r, lambda_ar = lambdas
    cctes_la, cctes_lr, cctes_2la, cctes_2lr, cctes_lar = cctes

    a1sb_a = a1sB(eta, I_la, J_la, lambda_a, cctes_la, eps)
    a1sb_r = a1sB(eta, I_lr, J_lr, lambda_r, cctes_lr, eps)
    a1sb_2a = a1sB(eta, I_2la, J_2la, 2*lambda_a, cctes_2la, eps)
    a1sb_2r = a1sB(eta, 2*lambda_r, cctes_2lr, eps)
    a1sb_ar = a1sB(eta, lambda_ar, cctes_lar, eps)

    a1sb_a1 = np.hstack([a1sb_a, a1sb_r])
    a1sb_a2 = np.hstack([a1sb_2a, a1sb_ar, a1sb_2r])
    return a1sb_a1, a1sb_a2


# Function for creating arrays with a1s + B and it's first derivative
# at all lambdas
def da1B_eval(eta, I_lambdas, J_lambdas, lambdas, cctes, eps):
    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdas
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdas
    lambda_a, lambda_r, lambda_ar = lambdas
    cctes_la, cctes_lr, cctes_2la, cctes_2lr, cctes_lar = cctes

    da1sb_a = da1sB(eta, I_la, J_la, lambda_a, cctes_la, eps)
    da1sb_r = da1sB(eta, I_lr, J_lr, lambda_r, cctes_lr, eps)
    da1sb_2a = da1sB(eta, I_2la, J_2la, 2*lambda_a, cctes_2la, eps)
    da1sb_2r = da1sB(eta, I_2lr, J_2lr, 2*lambda_r, cctes_2lr, eps)
    da1sb_ar = da1sB(eta, I_lar, J_lar, lambda_ar, cctes_lar, eps)

    a1sb_a1 = np.column_stack([da1sb_a, da1sb_r])
    a1sb_a2 = np.column_stack([da1sb_2a, da1sb_ar, da1sb_2r])
    return a1sb_a1, a1sb_a2


# Function for creating arrays with a1s + B and it's first
# and second derivative at all lambdas
def d2a1B_eval(eta, I_lambdas, J_lambdas, lambdas, cctes, eps):

    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdas
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdas
    lambda_a, lambda_r, lambda_ar = lambdas
    cctes_la, cctes_lr, cctes_2la, cctes_2lr, cctes_lar = cctes

    d2a1sb_a = d2a1sB(eta, I_la, J_la, lambda_a, cctes_la, eps)
    d2a1sb_r = d2a1sB(eta, I_lr, J_lr, lambda_r, cctes_lr, eps)
    d2a1sb_2a = d2a1sB(eta, I_2la, J_2la, 2*lambda_a, cctes_2la, eps)
    d2a1sb_2r = d2a1sB(eta, I_2lr, J_2lr, 2*lambda_r, cctes_2lr, eps)
    d2a1sb_ar = d2a1sB(eta,  I_lar, J_lar, lambda_ar, cctes_lar, eps)

    a1sb_a1 = np.column_stack([d2a1sb_a, d2a1sb_r])
    a1sb_a2 = np.column_stack([d2a1sb_2a, d2a1sb_ar, d2a1sb_2r])

    return a1sb_a1, a1sb_a2


# Function for creating arrays with a1s + B and it's first, second and third
# derivative at all lambdas
def d3a1B_eval(eta, I_lambdas, J_lambdas, lambdas, cctes, eps):

    I_la, I_lr, I_2la, I_2lr, I_lar = I_lambdas
    J_la, J_lr, J_2la, J_2lr, J_lar = J_lambdas
    lambda_a, lambda_r, lambda_ar = lambdas
    cctes_la, cctes_lr, cctes_2la, cctes_2lr, cctes_lar = cctes

    d3a1sb_a = d3a1sB(eta, I_la, J_la, lambda_a, cctes_la, eps)
    d3a1sb_r = d3a1sB(eta, I_lr, J_lr, lambda_r, cctes_lr, eps)
    d3a1sb_2a = d3a1sB(eta, I_2la, J_2la, 2*lambda_a, cctes_2la, eps)
    d3a1sb_2r = d3a1sB(eta, I_2lr, J_2lr, 2*lambda_r, cctes_2lr, eps)
    d3a1sb_ar = d3a1sB(eta,  I_lar, J_lar, lambda_ar, cctes_lar, eps)

    a1sb_a1 = np.column_stack([d3a1sb_a, d3a1sb_r])
    a1sb_a2 = np.column_stack([d3a1sb_2a, d3a1sb_ar, d3a1sb_2r])

    return a1sb_a1, a1sb_a2


def x0lambda_eval(x0, lambda_a, lambda_r, lambda_ar):
    x0la = x0**lambda_a
    x0lr = x0**lambda_r
    x02la = x0**(2*lambda_a)
    x02lr = x0**(2*lambda_r)
    x0lar = x0**lambda_ar

    x0_a1 = np.hstack([x0la, -x0lr])
    x0_a2 = np.hstack([x02la, -2*x0lar, x02lr])

    x0_a12 = np.hstack([lambda_a*x0la, -lambda_r*x0lr])
    x0_a22 = np.hstack([lambda_a*x02la, -lambda_ar*x0lar, lambda_r*x02lr])
    return x0_a1, x0_a2, x0_a12, x0_a22
